$ifndef _USB_CDC_INC_
$define _USB_CDC_INC_

$define cUSB_USE_CDC
'
' Class-Specific Requests
'
    Symbol cSEND_ENCAPSULATED_COMMAND  = $00
    Symbol cGET_ENCAPSULATED_RESPONSE  = $01
    Symbol cSET_COMM_FEATURE           = $02
    Symbol cGET_COMM_FEATURE           = $03
    Symbol cCLEAR_COMM_FEATURE         = $04
    Symbol cSET_LINE_CODING            = $20
    Symbol cGET_LINE_CODING            = $21
    Symbol cSET_CONTROL_LINE_STATE     = $22
    Symbol cSEND_BREAK                 = $23
'
' Notifications
'
' Notes: Notifications are polled over
' Communication Interface (Interrupt Endpoint)
'
    Symbol cNETWORK_CONNECTION         = $00
    Symbol cRESPONSE_AVAILABLE         = $01
    Symbol cSERIAL_STATE               = $20
'
' Device Class Code
'
    Symbol cCDC_DEVICE                 = $02
'
' Communication Interface Class Code
'
    Symbol cCOMM_INTF                  = $02
'
' Communication Interface Class SubClass Codes
'
    Symbol cABSTRACT_CONTROL_MODEL     = $02
'
' Communication Interface Class Control Protocol Codes
'
    Symbol cV25TER                     = $01    ' Common AT commands ("Hayes(TM)")
'
' Data Interface Class Codes
'
    Symbol cDATA_INTF                  = $0A
'
' Data Interface Class Protocol Codes
'
    Symbol cNO_PROTOCOL                = $00     ' No class specific protocol required
'
' Communication Feature Selector Codes
'
    Symbol cABSTRACT_STATE             = $01
    Symbol cCOUNTRY_SETTING            = $02
'
' Functional Descriptors
' Type Values for the bDscType Field
'
    Symbol cCS_INTERFACE               = $24
    Symbol cCS_ENDPOINT                = $25
'
' bDscSubType in Functional Descriptors
'
    Symbol cDSC_FN_HEADER              = $00
    Symbol cDSC_FN_CALL_MGT            = $01
    Symbol cDSC_FN_ACM                 = $02    ' ACM - Abstract Control Management
    Symbol cDSC_FN_DLM                 = $03    ' DLM - Direct Line Managment
    Symbol cDSC_FN_TELEPHONE_RINGER    = $04
    Symbol cDSC_FN_RPT_CAPABILITIES    = $05
    Symbol cDSC_FN_UNION               = $06
    Symbol cDSC_FN_COUNTRY_SELECTION   = $07
    Symbol cDSC_FN_TEL_OP_MODES        = $08
    Symbol cDSC_FN_USB_TERMINAL        = $09

    Symbol cDUMMY_LENGTH               = $08
    Symbol cLINE_CODING_LENGTH         = $07

    Include "USB_Dev.inc"                   ' Include the main USB routines

Asm-
'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CheckCDCRequest
' Input:           Non
' Output:          None

' Overview:        This routine checks the setup data packet to see if it knows how to handle it
' Notes:           None
'
__USB_CheckCDCRequest
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
'
' If request cUSB_Recipient is not an interface then Return
'
    Movf USB_aSetupPkt,w                    ' Recipient = RCPT_INTF?
    Andlw 0x1F                              ' Mask to lower 5 bits
    Sublw cRCPT_INTF
    Bnz __USB_CheckCDCRequestExit
'
' If request type is not class-specific then Return
'
    Movf USB_aSetupPkt,w                    ' RequestType = CLASS?
    Andlw 0x60                              ' Mask to proper bits
    Sublw (cCLASS << 5)
    Bnz __USB_CheckCDCRequestExit
'
' Interface ID must match interface numbers associated with CDC class,else Return
'
    Movlw cCDC_COMM_INTF_ID                 ' IntfID = cCDC_COMM_INTF_ID?
    Subwf USB_aSetupPkt + cUSB_bIntfID,w
    Bz __USB_CheckCDCRequest1
    Movlw cCDC_DATA_INTF_ID                 ' IntfID = cCDC_DATA_INTF_ID?
    Cpfseq USB_aSetupPkt + cUSB_bIntfID
__USB_CheckCDCRequestExit
    Return

__USB_CheckCDCRequest1
    Movlw cSEND_ENCAPSULATED_COMMAND        ' Request = cSEND_ENCAPSULATED_COMMAND?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckCDCRequest2
'
' Send the packet
'
    Movlw (USB_aDummyEncapCmdResp & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_aDummyEncapCmdResp >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movlw cDUMMY_LENGTH
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_CtrlTrfMem           ' Indicate RAM
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckCDCRequest2
    Movlw cGET_ENCAPSULATED_RESPONSE        ' Request = cGET_ENCAPSULATED_RESPONSE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra USBCheckCDCRequest3
'
' Populate USB_aDummyEncapCmdResp first.
'
    Movlw (USB_aDummyEncapCmdResp & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_aDummyEncapCmdResp >> 8) & 0xFF)
    Movwf USB_wSrcH
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
    Return

USBCheckCDCRequest3
    Movlw cSET_LINE_CODING                   ' Request = cSET_LINE_CODING?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckCDCRequest7
    Movlw cLINE_CODING_LENGTH
    Movwf USB_wOutCount
    Clrf USB_wOutCountH
    Movlw (USB_aLineCoding & 0xFF)
    Movwf USB_wDst
    Movlw ((USB_aLineCoding >> 8) & 0xFF)
    Movwf USB_wDstH
    Bsf USB_bOutFo,cUSB_Busy
    Return

__USB_CheckCDCRequest7
    Movlw cGET_LINE_CODING                   ' Request = cGET_LINE_CODING?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckCDCRequest8
    Movlw (USB_aLineCoding & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_aLineCoding >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movlw cLINE_CODING_LENGTH
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_CtrlTrfMem                   ' Indicate RAM
    Bsf USB_bInFo,cUSB_IncludeZero                    ' Include a trailing zero packet
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckCDCRequest8
    Movlw cSET_CONTROL_LINE_STATE            ' Request = cSET_CONTROL_LINE_STATE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckCDCRequest9
    Movff USB_aSetupPkt + cUSB_wValue,USB_bControlSignalBitmap
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckCDCRequest9
    Movlw cSEND_BREAK                        ' Request = cSEND_BREAK?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Return
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
    Return

' *** User API ****
'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __CDC_InitEP
' Input:           None
' Output:          None

' Overview:        __CDC_InitEP Initialises CDC endpoints,buffer descriptors,
'                  internal state-machine,and variables.
'                  It should be called after the USB host has sent out a
'                  SET_CONFIGURATION request.
'                  See __USB_StdSetCfgHandler() in usb_device.c for examples.
' Notes:           None
'
__CDC_InitEP
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
'
' Abstract line coding information
'
    Movlw (115200 & 0xFF)                       ' \
    Movwf USB_aLineCoding + cUSB_dwDTERate      ' |
    Movlw ((115200 >> 8) & 0xFF)                ' |
    Movwf USB_aLineCoding + (cUSB_dwDTERate + 1)' | Baud rate
    Movlw ((115200 >> 16) & 0xFF)               ' |
    Movwf USB_aLineCoding + (cUSB_dwDTERate + 2)' |
    Clrf USB_aLineCoding + (cUSB_dwDTERate + 3) ' /
    Clrf USB_aLineCoding + cUSB_bCharFormat     ' 1 stop bit
    Clrf USB_aLineCoding + cUSB_bParityType     ' No parity
    Movlw 8
    Movwf USB_aLineCoding + cUSB_bDataBits      ' 8-bits (5, 6, 7, 8,or 16)
'
' Do not have to init cUSB_CNT of IN pipes here.
' Reason:  Number of BYTEs to send to the host varies from one transaction to another.
' cUSB_CNT should equal the exact number of BYTEs to transmit for a given IN transaction.
' This number of BYTEs will only be known right before the data is sent.
'
' Enable and configure IN EndPoint
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cCDC_COMM_EP
    Movwf GEN                                   ' EndPoint number
    Movlw cUSB_IN_ENABLED | cUSB_HANDSHAKE_ENABLED | cUSB_DISALLOW_SETUP
    Movwf USB_wTemp                             ' Options
    Lfsr 1,USB_aCdcNotice                       ' Endpoint buffer
    Movlw cCDC_COMM_IN_EP_SIZE                  ' Endpoint size
    Rcall __USB_EnableEndpointIn
'
' Enable and configure OUT EndPoint
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cCDC_DATA_EP
    Movwf GEN                                   ' Endpoint number
    Movlw cUSB_OUT_ENABLED | cUSB_IN_ENABLED | cUSB_HANDSHAKE_ENABLED | cUSB_DISALLOW_SETUP
    Movwf USB_wTemp                             ' Options
    Lfsr 1,USB_aRXReport                        ' Endpoint buffer
    Movlw cCDC_DATA_OUT_EP_SIZE                 ' Endpoint size
    Rcall __USB_EnableEndpointOut
'
' Enable and configure IN EndPoint
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cCDC_DATA_EP
    Movwf GEN                                   ' Endpoint number
    Movlw cUSB_OUT_ENABLED | cUSB_IN_ENABLED | cUSB_HANDSHAKE_ENABLED | cUSB_DISALLOW_SETUP
    Movwf USB_wTemp                             ' Options
    Lfsr 1,USB_aTXReport                        ' Endpoint buffer
    Movlw cCDC_DATA_IN_EP_SIZE                  ' Endpoint size
    Bra __USB_EnableEndpointIn

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBSuspend
' Input:           None
' Output:          None

' Overview:        Call back that is invoked when a USB suspend is detected
' Notes:           None
'
__USB_CBSuspend
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBWakeFromSuspend
' Input:           None
' Output:          None

' Overview:        The host may put USB peripheral devices in low power suspend mode (by "sending" 3+ms of idle).
'                  Once in suspend mode,the host may wake the device back up by sending non-idle state signalling.
'                  This call back is invoked when a wakeup from USB suspend is detected.
' Notes:           None
'
__USB_CBWakeFromSuspend
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CB_SOF_Handler
' Input:           None
' Output:          None

' Overview:        The USB host sends out a SOF packet to full-speed devices every 1 ms.
'                  This interrupt may be useful for isochronous pipes.
'                  End designers should implement callback routine as necessary.
' Notes:           None
'
__USB_CB_SOF_Handler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBErrorHandler
' Input:           None
' Output:          None

' Overview:        The purpose of this callback is mainly for debugging during development.
'                  Check UEIR to see which error causes the interrupt.
' Notes:           None
'
__USB_CBErrorHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBCheckOtherReq
' Input:           None
' Output:          None

' Overview:        When SETUP packets arrive from the host,some firmware must process the request and respond appropriately to fulfill the request.
'                  Some of the SETUP packets will be for standard USB "chapter 9" (as in,fulfilling chapter 9 of the official USB specifications) requests,
'                  while others may be specific to the USB device class that is being implemented.
'                  For example,a HID class device needs to be able to respond to "GET REPORT" type of requests.
'                  This is not a standard USB chapter 9 request,and therefore not handled by usb_device.c.
'                  Instead this request should be handled by class specific firmware,such as that contained in usb_function_hid.c.
' Notes:           None
'
__USB_CBCheckOtherReq
    Bra __USB_CheckCDCRequest

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBStdSetDscHandler
' Input:           None
' Output:          None

' Overview:        The __USB_CBStdSetDscHandler() callback subroutine is called when a SETUP,cUSB_bRequest: SET_DESCRIPTOR request arrives.
'                  Typically SET_DESCRIPTOR requests are not used in most applications,and it is optional to support this type of request.
' Notes:           None
'
__USB_CBStdSetDscHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBInitEP
' Input:           None
' Output:          None

' Overview:        This subroutine is called when the device becomes initialised,which occurs after the host sends a SET_CONFIGURATION (cUSB_wValue not = 0) request.
'                  This callback subroutine should Initialise the endpoints for the device's usage according to the current configuration.
' Notes:           None
'
__USB_CBInitEP
    Bra __CDC_InitEP

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:      __USB_CBSendResume
' Input:           None
' Output:          None

' Overview:        The USB specifications allow some types of USB peripheral devices to wake up a host PC (such as if it is in a low power suspend to RAM state).
'                  This can be a very useful feature in some USB applications,such as an Infrared remote control receiver.
'                  If a user presses the "power" button on a remote control,it is nice that the IR receiver can detect this signalling,
'                  and then send a USB "command" to the PC to wake up.
'
'                  The __USB_CBSendResume() "callback" subroutine is used to send this special USB signalling which wakes up the PC.
'                  This subroutine may be called by application firmware to wake up the PC.
'                  This subroutine should only be called when:
'
'                    1.  The USB driver used on the host PC supports the remote wakeup capability.
'                    2.  The USB configuration descriptor indicates the device is remote wakeup capable in the bmAttributes field.
'                    3.  The USB host PC is currently sleeping, and has previously sent your device a SET FEATURE setup packet which "armed" the remote wakeup capability.
'
'                  This callback should send a RESUME signal that has the period of 1-15ms.
'
' Notes:           Interrupt vs. Polling
'                  -Primary clock
'                  -Secondary clock ***** MAKE NOTES ABOUT THIS *******
'                   > Can switch to primary first by calling __USB_CBWakeFromSuspend()

'                  The modifiable section in this routine should be changed to meet the application needs.
'                  Current implementation temporary blocks other functions from executing for a period of 1-13 ms depending on the core frequency.
'
'                  According to USB 2.0 specification section 7.1.7.7, "The remote wakeup device must hold the resume signaling for at lest 1 ms but for no more than 15 ms."
'                  The idea here is to use a delay counter loop,using a common value that would work over a wide range of core frequencies.
'                  That value selected is 1800. See table below:
'                  ==========================================================
'                  Core Freq(MHz)      MIP         RESUME Signal Period (ms)
'                  ==========================================================
'                      48              12          1.05
'                       4              1           12.6
'                  ==========================================================
'
__USB_CBSendResume
    Movlb (_usb_ram >> 8)           ' Point to the USB RAM bank
    Bsf UCON,Resume                 ' Start RESUME signaling
    Movlw 0x10                      ' Set RESUME line for 1-13 ms
    Movwf FSR2H
    Clrf FSR2L
USBCBSendResumeLoop
    Decfsz FSR2L,f
    Bra USBCBSendResumeLoop
    Decfsz FSR2H,f
    Bra USBCBSendResumeLoop
    Bcf UCON,Resume
    Return
Endasm-

$endif ' _USB_CDC_INC_
