$ifndef _USB_DEV_INC_
$define _USB_DEV_INC_
'
' USB routines
'
' Buffer Descriptor Status Register Initialisation Parameters
'
    Symbol _cBSTALL      = $04        ' Buffer Stall enable
    Symbol _cDTSEN       = $08        ' Data Toggle Synch enable
    Symbol _cINCDIS      = $10        ' Address increment disable
    Symbol _cKEN         = $20        ' SIE keeps buff descriptors enable
    Symbol _cDAT0        = $00        ' DATA0 packet expected next
    Symbol _cDAT1        = $40        ' DATA1 packet expected next
    Symbol _cDTSMASK     = $40        ' cUSB_DTS Mask
    Symbol _cUSIE        = $80        ' SIE owns buffer
    Symbol _cUCPU        = $00        ' CPU owns buffer
'
' EndPoint Address Format: X:EP3:EP2:EP1:EP0:DIR:PPBI:X
' This is used when checking the value read from USTAT
' Notes: These definitions are not used in the descriptors.
' EP addresses used in the descriptors have different format.
'
    Symbol cUSTAT_EP0_PP_MASK   = $FD ' ~= $02
    Symbol cUSTAT_EP_MASK       = $7E
    Symbol cUSTAT_EP0_OUT       = $00
    Symbol cUSTAT_EP0_IN        = $04

    Symbol cUEP_STALL           = $01
'
' USB EndPoint Definitions
' USB Standard EP Address Format: DIR:X:X:X:EP3:EP2:EP1:EP0
' This is used in the descriptors.
' NOTE: Do not use these values for checking against USTAT.
'
    Symbol _cEP_IN  = $80
    Symbol _cEP_OUT = $00
'
' Configuration Attributes
'
    Symbol _cDEFAULT = 128    ' Default Value (Bit 7 is set)
    Symbol _cSELF    = 64     ' Self-powered (Supports if set)
    Symbol _cRWU     = 32     ' Remote Wakeup (Supports if set)
'
' EndPoint Transfer Type
'
    Symbol _cCTRL       = $00        ' Control Transfer
    Symbol _cISO        = $01        ' Isochronous Transfer
    Symbol _cBULK       = $02        ' Bulk Transfer
    Symbol _cINT        = $03        ' Interrupt Transfer
'
' Various options for setting the PIPES
'
    Symbol cUSB_INPIPES_ROM          = $00    ' Data comes from RAM
    Symbol cUSB_INPIPES_RAM          = $01    ' Data comes from ROM
    Symbol cUSB_INPIPES_BUSY         = $80    ' The PIPE is cUSB_Busy
    Symbol cUSB_INPIPES_INCLUDE_ZERO = $40    ' #Include a trailing zero packet
    Symbol cUSB_INPIPES_NO_DATA      = $00    ' No data to send
    Symbol cUSB_INPIPES_NO_OPTIONS   = $00    ' No options set

    Symbol cUSB_EP0_ROM              = cUSB_INPIPES_ROM
    Symbol cUSB_EP0_RAM              = cUSB_INPIPES_RAM
    Symbol cUSB_EP0_BUSY             = cUSB_INPIPES_BUSY
    Symbol cUSB_EP0_INCLUDE_ZERO     = cUSB_INPIPES_INCLUDE_ZERO
    Symbol cUSB_EP0_NO_DATA          = cUSB_INPIPES_NO_DATA
    Symbol cUSB_EP0_NO_OPTIONS       = cUSB_INPIPES_NO_OPTIONS
'
' Standard Request Codes
' USB 2.0 Spec Ref Table 9-4
'
    Symbol cGET_STATUS   = 0
    Symbol cCLR_FEATURE  = 1
    Symbol cSET_FEATURE  = 3
    Symbol cSET_ADR      = 5
    Symbol cGET_DSC      = 6
    Symbol cSET_DSC      = 7
    Symbol cGET_CFG      = 8
    Symbol cSET_CFG      = 9
    Symbol cGET_INTF     = 10
    Symbol cSET_INTF     = 11
    Symbol cSYNCH_FRAME  = 12
'
' Standard Feature Selectors
'
    Symbol cDEVICE_REMOTE_WAKEUP = $01
    Symbol cENDPOINT_HALT        = $00
'
' USB Device States - To be used with [USB_bDeviceState]
'
    Symbol cDETACHED_STATE      = $00
    Symbol cATTACHED_STATE      = $01
    Symbol cPOWERED_STATE       = $02
    Symbol cDEFAULT_STATE       = $04
    Symbol cADR_PENDING_STATE   = $08
    Symbol cADDRESS_STATE       = $10
    Symbol cCONFIGURED_STATE    = $20

    Symbol cUSB_PING_PONG__NO_PING_PONG    = $00    ' %00000000
    Symbol cUSB_PING_PONG__EP0_OUT_ONLY    = $01    ' %00000001
    Symbol cUSB_PING_PONG__FULL_PING_PONG  = $02    ' %00000010
    Symbol cUSB_PING_PONG__ALL_BUT_EP0     = $03    ' %00000011
'
' UCFG Initialisation Parameters
'
    Symbol cUSB_PULLUP_ENABLE        = $10
    Symbol cUSB_PULLUP_DISABLED      = $00

    Symbol cUSB_INTERNAL_TRANSCEIVER = $00
    Symbol cUSB_EXTERNAL_TRANSCEIVER = $08

    Symbol cUSB_FULL_SPEED           = $04
    Symbol cUSB_LOW_SPEED            = $00

$defeval cUCFG_VAL cUSB_PULLUP_ENABLE | cUSB_INTERNAL_TRANSCEIVER | cUSB_FULL_SPEED | cUSB_PING_PONG__NO_PING_PONG
'
' UEPn Initialisation Parameters
'
    Symbol cEP_CTRL      = $06        ' Cfg Control pipe for this ep
    Symbol cEP_OUT       = $0C        ' Cfg OUT only pipe for this ep
    Symbol cEP_IN        = $0A        ' Cfg IN only pipe for this ep
    Symbol cEP_OUT_IN    = $0E        ' Cfg both OUT & IN pipes for this ep
'
' Handshake should be disable for isoch
'
    Symbol cUSB_HANDSHAKE_ENABLED   = $10
    Symbol cUSB_HANDSHAKE_DISABLED  = $00

    Symbol cUSB_OUT_ENABLED         = $04
    Symbol cUSB_OUT_DISABLED        = $00

    Symbol cUSB_IN_ENABLED          = $02
    Symbol cUSB_IN_DISABLED         = $00

    Symbol cUSB_ALLOW_SETUP         = $00
    Symbol cUSB_DISALLOW_SETUP      = $08

    Symbol cUSB_STALL_ENDPOINT      = $01
'
' Control Transfer States
'
    Symbol cWAIT_SETUP   = 0
    Symbol cCTRL_TRF_TX  = 1
    Symbol cCTRL_TRF_RX  = 2
'
' Short Packet States - Used by Control Transfer Read  - cCTRL_TRF_TX
'
    Symbol cSHORT_PKT_NOT_USED = 0
    Symbol cSHORT_PKT_PENDING  = 1
    Symbol cSHORT_PKT_SENT     = 2
'
' USB cUSB_PID: Token Types - See chapter 8 in the USB specification
'
    Symbol cSETUP_TOKEN  = $0D        ' %00001101
    Symbol cOUT_TOKEN    = $01        ' %00000001
    Symbol cIN_TOKEN     = $09        ' %00001001
'
' cUSB_bmRequestType Definitions
'
    Symbol cHOST_TO_DEV = 0
    Symbol cDEV_TO_HOST = 1

    Symbol cSTANDARD    = $00
    Symbol cCLASS       = $01
    Symbol cVENDOR      = $02

    Symbol cRCPT_DEV    = 0
    Symbol cRCPT_INTF   = 1
    Symbol cRCPT_EP     = 2
    Symbol cRCPT_OTH    = 3
'
' Section: USB Descriptors
'
    Symbol cUSB_DESCRIPTOR_DEVICE           = $01       ' cUSB_bDescriptorType for a Device Descriptor.
    Symbol cUSB_DESCRIPTOR_CONFIGURATION    = $02       ' cUSB_bDescriptorType for a Configuration Descriptor.
    Symbol cUSB_DESCRIPTOR_STRING           = $03       ' cUSB_bDescriptorType for a String Descriptor.
    Symbol cUSB_DESCRIPTOR_INTERFACE        = $04       ' cUSB_bDescriptorType for an Interface Descriptor.
    Symbol cUSB_DESCRIPTOR_ENDPOINT         = $05       ' cUSB_bDescriptorType for an EndPoint Descriptor.
    Symbol cUSB_DESCRIPTOR_DEVICE_QUALIFIER = $06       ' cUSB_bDescriptorType for a Device Qualifier.
    Symbol cUSB_DESCRIPTOR_OTHER_SPEED      = $07       ' cUSB_bDescriptorType for a Other Speed Configuration.
    Symbol cUSB_DESCRIPTOR_INTERFACE_POWER  = $08       ' cUSB_bDescriptorType for Interface Power.
    Symbol cUSB_DESCRIPTOR_OTG              = $09       ' cUSB_bDescriptorType for an OTG Descriptor.

    Include "USB_MemAlloc.inc"                          ' Include variables used by USB
Asm-
'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_DeviceInit
' Description   : This subroutine Initialises the device stack it in the default state
' Parameters    : None
' Output        : None
' Notes         : The USB module will be completely reset including all of the internal variables,registers,and interrupt flags.
'
__USB_DeviceInit
'
' Clear all USB error flags
'
NOLIST
#if(UEIR < 0x0F60)
LIST
    Movlb (UEIR >> 8)
NOLIST
#endif
LIST
    Clrf UEIR                               ' Clear all USB error flags
'
' Clear all USB interrupts
'
    Clrf UIR                                ' Clear all USB interrupts
NOLIST
#if(UEIE < 0x0F60)
LIST
    Movlb (UEIE >> 8)
NOLIST
#endif
LIST
    Movlw 0x9F                              ' Unmask all USB error interrupts
    Movwf UEIE
    Movlw 0xFB                              ' Enable all interrupts except ACTVIE
    Movwf UIE
    Clrf UADDR                              ' Reset to default address
'
' Clear all of the EndPoint control registers
'
    Lfsr 2,UEP1                             ' Clear all of the EndPoint control registers
    Movlw cUSB_MAX_EP_NUMBER
__USB_DeviceInitEPLoop
    Clrf POSTINC2
    Decfsz WREG,f
    Bra __USB_DeviceInitEPLoop
'
' Clear all of the BDT entries
'
    Movlw cEP_CTRL | cUSB_HANDSHAKE_ENABLED ' Initialise EP0 as a Ctrl EP
    Movwf UEP0
'
' Flush any pending transactions
'
    Bra __USB_DeviceInitFlush1
__USB_DeviceInitFlushLoop
    Bcf UIR,PP_TRNIF                        ' Flush any pending transactions
    Nop
    Nop
    Nop
    Nop
    Nop
    Nop
__USB_DeviceInitFlush1
    Btfsc UIR,PP_TRNIF
    Bra __USB_DeviceInitFlushLoop
'
' Clear all of the internal pipe information
'
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Clrf USB_bInFo
    Clrf USB_bOutFo
    Clrf USB_wOutCount
    Clrf USB_wOutCountH
    Clrf USB_bRemoteWakeup
    Bcf UCON,PKTDIS                         ' Make sure packet processing is enabled
'
' Get ready for the first packet
'
    Clrf USB_bActiveConfiguration            ' Clear active configuration
    Movlw cDETACHED_STATE                    ' Indicate that we are now in the detached state
    Movwf USB_bDeviceState
    Movlb 0
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_DeviceTasks

' Description:  This subroutine is the main state machine of the USB device side stack.
'               This subroutine should be called periodically to receive and transmit packets through the stack.
'               This subroutine should be called  preferably once every 100us during the enumeration process.
'               After the enumeration process this subroutine still needs to be called periodically to respond to various
'               situations on the bus but is more relaxed in its time requirements.
'               This subroutine should also be called at least as fast as the OUT data expected from the PC.
' Output:       None
' Notes:        None
'
__USB_DeviceTasks
Endasm-
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
$ifdef cUSE_USB_BUS_SENSE_IO
    Btfsc USB_BUS_SENSE_PIN                 ' If we aren't attached to the bus
    Bra __USB_DeviceTasksCheckDetached
    Clrf UCON                               ' Disable module & detach from bus
#if(UIE < 0x0F60)
    Movlb (UIE >> 8)
#endif
    Clrf UIE                                ' Mask all USB interrupts
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Movlw cDETACHED_STATE                   ' Move to the detached state
    Movwf USB_bDeviceState
    Return                                  ' Return so that we don't go through the rest of the state machine
$endif

Asm-
__USB_DeviceTasksCheckDetached
'
' If we are in the detached state
'
    Movlw cDETACHED_STATE
    Cpfseq USB_bDeviceState
    Bra __USB_DeviceTasksCheckAttached
    Clrf UCON                               ' Disable module & detach from bus
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Clrf UIE                                ' Mask all USB interrupts
__USB_DeviceTasksCDLoop
    Bsf UCON,USBEN                          ' Enable module & attach to bus
    Btfss UCON,USBEN
    Bra __USB_DeviceTasksCDLoop
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Movlw cATTACHED_STATE
    Movwf USB_bDeviceState
NOLIST
#if(UCFG < 0x0F60)
LIST
    Movlb (UCFG >> 8)
NOLIST
#endif
LIST
    Movlw cUCFG_VAL                          ' Enable/set things like: pull ups,full/low-speed mode...
    Movwf UCFG

__USB_DeviceTasksCheckAttached
    Movlw cATTACHED_STATE
    Cpfseq USB_bDeviceState
    Bra __USB_DeviceTasksA
'
' After enabling the USB module,it takes some time for the voltage on the D+ or D- line to rise high enough to get out of the SE0 condition.
' The USB Reset interrupt should not be unmasked until the SE0 condition is cleared.
' This helps prevent the firmware from misinterpreting this unique event as a USB bus reset from the USB host.
'
    Btfsc UCON,SE0
    Bra __USB_DeviceTasksA
    Clrf UIR                                ' Clear all USB interrupts
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Clrf UIE                                ' Mask all USB interrupts
    Bsf UIE,URSTIE                          ' Unmask RESET interrupt
    Bsf UIE,IDLEIE                          ' Unmask IDLE interrupt
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Movlw cPOWERED_STATE
    Movwf USB_bDeviceState
'
' Task A: Service USB Activity Interrupt
'
__USB_DeviceTasksA
    Btfss UIR,ACTVIF
    Bra __USB_DeviceTasks1
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfsc UIE,ACTVIE
    Rcall __USB_WakeFromSuspend
'
' Task B: Service USB Bus Reset Interrupt.
' When bus reset is received during suspend, ACTVIF will be set first, once the UCONbits.SUSPND is clear,then the URSTIF bit will be asserted.
' This is why URSTIF is checked after ACTVIF.
'
__USB_DeviceTasks1
    Btfsc UCON,SUSPND                       ' Are we suspended?
    Return                                  ' Pointless to continue servicing if the device is in suspend mode.
'
' The USB reset flag is masked when the USB state is in DETACHED_STATE or ATTACHED_STATE,
' and therefore cannot cause a USB reset event during these two states.
'
    Btfss UIR,URSTIF                        ' USB Bus Reset Interrupt?
    Bra __USB_DeviceTasks2
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfss UIE,URSTIE
    Bra __USB_DeviceTasks2
    Rcall __USB_DeviceInit
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Movlw cDEFAULT_STATE
    Movwf USB_bDeviceState
'
' Bug Fix: Feb 26,2007 v2.1 (#F1)
'
' In the original firmware,if an OUT token is sent by the host before a SETUP token is sent,the firmware would respond with an ACK.
' This is not a correct response,the firmware should have sent a STALL.
' This is a minor non-compliance since a compliant host should not send an OUT before sending a SETUP token.
' The fix allows a SETUP transaction to be accepted while stalling OUT transactions.
'
    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw _cUSIE | _cDAT0 | _cDTSEN | _cBSTALL
    Movwf USB_dEp0Bo + cUSB_STAT
'
' Task C: Service other USB interrupts
'
__USB_DeviceTasks2
    Btfss UIR,IDLEIF
    Bra __USB_DeviceTasks3
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfss UIE,IDLEIE
    Bra __USB_DeviceTasks3
    Rcall __USB_Suspend
    Bcf UIR,IDLEIF

__USB_DeviceTasks3
    Btfss UIR,SOFIF
    Bra __USB_DeviceTasks4

NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfss UIE,SOFIE
    Bra __USB_DeviceTasks4
    Rcall __USB_CB_SOF_Handler
    Bcf UIR,SOFIF

__USB_DeviceTasks4
    Btfss UIR,STALLIF
    Bra __USB_DeviceTasks5
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfsc UIE,STALLIE
    Rcall __USB_StallHandler

__USB_DeviceTasks5
    Btfss UIR,UERRIF
    Bra __USB_DeviceTasks6
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfss UIE,UERRIE
    Bra __USB_DeviceTasks6
    Rcall __USB_CBErrorHandler
    Clrf UIR                                    ' Clear UERRIF

__USB_DeviceTasks6
'
' Pointless to continue servicing if the host has not sent a bus reset.
' Once bus reset is received,the device transitions into the DEFAULT state and is ready for communication.
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cDEFAULT_STATE
    Subwf USB_bDeviceState,w
    Bnc __USB_DriverTasksExit
'
' Task D: Servicing USB Transaction Complete Interrupt
'
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Btfss UIE,TRNIE
    Return
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw 4                                     ' Drain or deplete the USAT FIFO entries.  If the USB FIFO ever gets full,USB bandwidth
    Movwf USB_bTRNIFCount                       ' utilisation can be compromised,and the device won't be able to receive SETUP packets.
__USB_DeviceTasks7
    Btfss UIR,PP_TRNIF
    Return
    Movf USTAT,w
    Movwf USB_bUSTATcopy
    Bcf UIR,PP_TRNIF
'
' __USB_CtrlEPService only services transactions over EP0.
' It ignores all other EP transactions.
'
    Rcall __USB_CtrlEPService
    Decfsz USB_bTRNIFCount,f
    Bra __USB_DeviceTasks7
__USB_DriverTasksExit
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_StallHandler
' Input:        None
' Output:       None
' Overview:     This subroutine handles the event of a STALL occuring on the bus
' Notes:         None
'
__USB_StallHandler
'
' Does not really have to do anything here, even for the control EndPoint.
' All BDs of EndPoint 0 are owned by SIE right now, but once a Setup Transaction is received, the ownership for EP0_OUT will be Returned to CPU.
' When the Setup Transaction is serviced,the ownership for EP0_IN will then be forced back to CPU by firmware.
'
NOLIST
#if(UEP0 < 0x0F60)
LIST
    Movlb (UEP0 >> 8)
NOLIST
#endif
LIST
    Btfss UEP0,EPSTALL
    Bra __USB_StallHandlerExit
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw _cUSIE                                ' USB_UOWN - [0]owned by CPU [1]owned by SIE
    Cpfseq USB_dEp0Bo + cUSB_STAT
    Bra __USB_StallHandlerExit
    Movlw _cUSIE | _cBSTALL
    Cpfseq USB_dEp0Bi + cUSB_STAT
    Bra __USB_StallHandlerExit
    Movlw _cUSIE | _cDAT0 | _cDTSEN | _cBSTALL  ' Set USB_dEp0Bo to stall also
    Movwf USB_dEp0Bo + cUSB_STAT
NOLIST
#if(UEP0 < 0x0F60)
LIST
    Movlb (UEP0 >> 8)
NOLIST
#endif
LIST
    Bcf UEP0,EPSTALL                             ' Clear stall status

__USB_StallHandlerExit
    Bcf UIR,STALLIF
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_Suspend
' Input:        None
' Output:       None
' Overview:     This subroutine handles if the host tries to suspend the device
' Notes:        None
'
__USB_Suspend
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Bsf UIE,ACTVIE                              ' Enable bus activity interrupt
    Bcf UIR,IDLEIF
    Bsf UCON,SUSPND                             ' Put USB module in power conserve
'
' At this point the PIC can go into sleep,idle,or switch to a slower clock,etc.
' This should be done in the __USB_CBSuspend() if necessary.
'
    Bra __USB_CBSuspend

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_WakeFromSuspend
' Input:        None
' Output:       None
' Notes:        None
'
__USB_WakeFromSuspend
    Bcf UCON,SUSPND                             ' Bring USB module out of power conserve mode
'
' If using clock switching,the place to restore the original microcontroller core clock frequency is in the __USB_CBWakeFromSuspend() callback
'
    Rcall __USB_CBWakeFromSuspend
NOLIST
#if(UIE < 0x0F60)
LIST
    Movlb (UIE >> 8)
NOLIST
#endif
LIST
    Bcf UIE,ACTVIE
'
' Bug Fix: Feb 26,2007 v2.1
'
' The ACTVIF bit cannot be cleared immediately after the USB module wakes up from Suspend or while the USB module is suspended.
' A few clock cycles are required to synchronize the internal hardware state machine before the ACTIVIF bit can be cleared by firmware.
' Clearing the ACTVIF bit before the internal hardware is synchronised may not have an effect on the value of ACTVIF.
' Additonally,if the USB module uses the clock from the 96 MHz PLL source,then after clearing the SUSPND bit,the USB
' module may not be immediately operational while waiting for the 96 MHz PLL to lock.
'
__USB_WakeFromSuspendLoop
    Btfss UIR,ACTVIF
    Return
    Bcf UIR,ACTVIF
    Bra __USB_WakeFromSuspendLoop

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:   __USB_CtrlEPService
' PreCondition: USTAT is loaded with a valid EndPoint address.
' Input:        None
' Output:       None
' Overview:     __USB_CtrlEPService checks for three transaction types that it knows how to service and services them:
'               1. EP0 SETUP
'               2. EP0 OUT
'               3. EP0 IN
'               It ignores all other types (i.e. EP1,EP2,etc.)
' Notes:        None
'
__USB_CtrlEPService
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
'
' If the last packet was a EP0 OUT packet
'
    Movf USB_bUSTATcopy,w
    Andlw cUSTAT_EP0_PP_MASK
    Sublw cUSTAT_EP0_OUT
    Bnz __USB_CtrlEPService1
    Movf USB_dEp0Bo + cUSB_STAT,w
    Andlw 0x3C                                  ' Mask to USB_PID
    Sublw (cSETUP_TOKEN << 2)
    Bz __USB_CtrlTrfSetupHandler
    Bra __USB_CtrlTrfOutHandler
__USB_CtrlEPService1
    Movf USB_bUSTATcopy,w
    Andlw cUSTAT_EP0_PP_MASK
    Sublw cUSTAT_EP0_IN
    Bz __USB_CtrlTrfInHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlTrfSetupHandler
' PreCondition:     USB_aSetupPkt buffer is loaded with valid USB Setup Data
' Input:            None
' Output:           None
' Overview:         This routine is a task dispatcher and has 3 stages.
'                       1. It Initialises the control transfer state machine.
'                       2. It calls on each of the module that may know how to service the Setup Request from the host.
'                           Module Example: USBD,HID,CDC,MSD,...
'                           A callback subroutine,__USB_CBCheckOtherReq(), is required to call other module handlers.
'                       3. Once each of the modules has had a chance to check if it is responsible for servicing the request,stage 3
'                          then checks direction of the transfer to determine how to prepare EP0 for the control transfer.
'                   Refer to __USB_CtrlEPServiceComplete() for more details.
' Notes:            Microchip USB Firmware has three different states for the control transfer state machine:
'                       1. cWAIT_SETUP
'                       2. cCTRL_TRF_TX
'                       3. cCTRL_TRF_RX
'                   Refer to firmware manual to find out how one state is transitioned to another.
'
'                   A Control Transfer is composed of many USB transactions.
'                   When transferring data over multiple transactions, it is important to keep track of data source,data destination,and data count.
'                   These three parameters are stored in USB_wSrc,USB_wDst,and wCount.
'                   A flag is used to note if the data source is from ROM or RAM.
'
__USB_CtrlTrfSetupHandler
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw _cUCPU
    Btfsc USB_dEp0Bi + cUSB_STAT,cUSB_UOWN
    Movwf USB_dEp0Bi + cUSB_STAT
'
' Keep track of if a short packet has been sent yet or not
'
    Movlw cSHORT_PKT_NOT_USED
    Movwf USB_bShortPacketStatus
'
' Stage 1
'
    Movlw cWAIT_SETUP
    Movwf USB_bControlTransferState
    Clrf USB_wInCount
    Clrf USB_wInCountH
    Clrf USB_bInFo
'
' Stage 2
'
    Rcall __USB_CheckStdRequest
    Rcall __USB_CBCheckOtherReq                    ' Class request
'
' Stage 3
'
    Bra __USB_CtrlEPServiceComplete

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlTrfOutHandler
' Input:            None
' Output:           None

' Overview:         This routine handles an OUT transaction according to which control transfer state is currently active.
' Notes:            If the the control transfer was from host to device,the session owner should be notified
'                   at the end of each OUT transaction to service the received data.
'
__USB_CtrlTrfOutHandler
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cCTRL_TRF_RX
    Cpfseq USB_bControlTransferState
    Bra __USB_PrepareForNextSetupTrf
    Bra __USB_CtrlTrfRxService

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlTrfInHandler
' Input:            None
' Output:           None

' Overview:         This routine handles an IN transaction according to which control transfer state is currently active.
' Notes:            A Set Address Request must not change the acutal addressof the device until the completion of the control transfer.
'                   The end of the control transfer for Set Address Request is an IN transaction.
'                   Therefore it is necessary to service this unique situation when the condition is right.
'
__USB_CtrlTrfInHandler
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Clrf USB_bLastDTS
    Btfsc USB_dEp0Bi + cUSB_STAT,cUSB_DTS
    Setf USB_bLastDTS
'
' Switch to the next ping pong buffer
'
    Movlw cADR_PENDING_STATE                    ' Must check if in ADR_PENDING_STATE
    Cpfseq USB_bDeviceState
    Bra __USB_CtrlTrfInHandler1
    Movf USB_aSetupPkt + cUSB_bDevADR,w
NOLIST
#if(UADDR < 0x0F60)
LIST
    Movlb (UADDR >> 8)
NOLIST
#endif
LIST
    Movwf UADDR
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cADDRESS_STATE                        ' If UADDR > 0
    Btfsc STATUS,Z                              ' Set from above
    Movlw cDEFAULT_STATE
    Movwf USB_bDeviceState

__USB_CtrlTrfInHandler1
    Movlw cCTRL_TRF_TX
    Cpfseq USB_bControlTransferState
    Bra __USB_PrepareForNextSetupTrf
    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_dEp0Bi + cUSB_ADRL
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_dEp0Bi + cUSB_ADRH
    Rcall __USB_CtrlTrfTxService
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cSHORT_PKT_SENT
    Cpfseq USB_bShortPacketStatus
    Bra __USB_CtrlTrfInHandler2
    Movlw _cUSIE | _cBSTALL                     ' If a short packet has been sent,don't want to send any more,
    Movwf USB_dEp0Bi + cUSB_STAT                ' stall next time if host is still trying to read.
    Return

__USB_CtrlTrfInHandler2
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Tstfsz USB_bLastDTS
    Movlw _cUSIE | _cDAT0 | _cDTSEN
    Movwf USB_dEp0Bi + cUSB_STAT
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_PrepareForNextSetupTrf
' Input:            None
' Output:           None

' Overview:        The routine forces EP0 OUT to be ready for a new Setup transaction,and forces EP0 IN to be owned by CPU.
' Notes:            None
'
__USB_PrepareForNextSetupTrf
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
'
' Bug Fix: Feb 26,2007 v2.1
'
' Facts:
' A Setup Packet should never be stalled. (USB 2.0 Section 8.5.3)
' If a Setup PID is detected by the SIE,the DTSEN setting is ignored.
' This causes a problem at the end of a control write transaction.
' In __USB_CtrlEPServiceComplete(),during a control write (Host to Device), the EP0_OUT is setup to write any data to the USB_aCtrlTrfData buffer.
' If <SETUP[0]><IN[1]> is completed and __USB_CtrlTrfInHandler() is not called before the next <SETUP[0]> is received,
' then the latest Setup data will be written to the USB_aCtrlTrfData buffer instead of the USB_aSetupPkt buffer.
'
' If __USB_CtrlTrfInHandler() was called before the latest <SETUP[0]> is received,then there would be no problem,
' because __USB_PrepareForNextSetupTrf() would have been called and updated USB_dEp0Bo.ADR to point to the USB_aSetupPkt buffer.
'
' Work around:
' Check for the problem as described above and copy the Setup data from USB_aCtrlTrfData to USB_aSetupPkt.
'
    Movlw cCTRL_TRF_RX
    Cpfseq USB_bControlTransferState
    Bra __USB_PrepForNextSetupTrf1
    Btfss UCON,PKTDIS
    Bra __USB_PrepForNextSetupTrf1
    Movlw cUSB_EP0_BUFF_SIZE
    Cpfseq USB_dEp0Bo + cUSB_CNT
    Bra __USB_PrepForNextSetupTrf1
    Movf USB_dEp0Bo + cUSB_STAT,w
    Andlw 0x3C                                  ' Mask to PID
    Sublw (cSETUP_TOKEN) << 2
    Bnz __USB_PrepForNextSetupTrf1
    Btfsc USB_dEp0Bo + cUSB_STAT,cUSB_UOWN
    Bra __USB_PrepForNextSetupTrf1

    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
'
' The Setup data was written to the USB_aCtrlTrfData buffer,
' must copy it back to the USB_aSetupPkt buffer so that it can be processed correctly by __USB_CtrlTrfSetupHandler().
'
    Lfsr 1,USB_aCtrlTrfData
    Lfsr 2,USB_aSetupPkt
    Movlw cUSB_EP0_BUFF_SIZE
__USB_PrepForNextSetupTrfLoop
    Movff POSTINC1,POSTINC2
    Decfsz WREG,f
    Bra __USB_PrepForNextSetupTrfLoop
    Bra __USB_PrepForNextSetupTrf2

__USB_PrepForNextSetupTrf1
    Movlw cWAIT_SETUP
    Movwf USB_bControlTransferState
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
'
' Bug Fix: Feb 26,2007 v2.1 (#F1)
'
' In the original firmware,if an OUT token is sent by the host before a SETUP token is sent,the firmware would respond with an ACK.
' This is not a correct response,the firmware should have sent a STALL.
' This is a minor non-compliance since a compliant host should not send an OUT before sending a SETUP token.
' The fix allows a SETUP transaction to be accepted while stalling OUT transactions.
'
    Movlw _cUSIE | _cDAT0 | _cDTSEN | _cBSTALL
    Movwf USB_dEp0Bo + cUSB_STAT
'
' Bug Fix: Feb 26,2007 v2.1 (#F3)
'
' In the original firmware,if an IN token is sent by the host before a SETUP token is sent,the firmware would respond with an ACK.
' This is not a correct response,the firmware should have sent a STALL.
' This is a minor non-compliance since a compliant host should not send an IN before sending a SETUP token.
'
    Movlw _cUCPU                                 ' EP0 IN buffer initialisation
    Movwf USB_dEp0Bi + cUSB_STAT

__USB_PrepForNextSetupTrf2
'
' If someone is still expecting data from the control transfer,
' then make sure to terminate that request and let them know that they are done
'
    Bcf USB_bOutFo,cUSB_Busy
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CheckStdRequest
' Input:            None
' Output:           None
' Overview:         This routine checks the setup data packet to see if it knows how to handle it
' Notes:            None
'
__USB_CheckStdRequest
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movf USB_aSetupPkt,w                        ' RequestType = STANDARD?
    Andlw 0x60                                  ' Mask to proper bits
    Sublw (cSTANDARD) << 5
    Bnz __USB_CheckStdRequestExit
    Movlw cSET_ADR                              ' Handle request
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest1
    Bsf USB_bInFo,cUSB_Busy
    Movlw cADR_PENDING_STATE                    ' Update state only
    Movwf USB_bDeviceState
__USB_CheckStdRequestExit
    Return

__USB_CheckStdRequest1
    Movlw cGET_DSC
    Subwf USB_aSetupPkt + cUSB_bRequest,w
    Bz __USB_StdGetDscHandler
    Movlw cSET_CFG
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest3
    Bra __USB_StdSetCfgHandler

__USB_CheckStdRequest3
    Movlw cGET_CFG
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest4

    Movlw (USB_bActiveConfiguration & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_bActiveConfiguration >> 8) & 0xFF)
    Movwf USB_wSrcH

    Bsf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate RAM
    Movlw 1
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckStdRequest4
    Movlw cGET_STATUS
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest5
    Bra __USB_StdGetStatusHandler

__USB_CheckStdRequest5
    Movlw cCLR_FEATURE
    Subwf USB_aSetupPkt + cUSB_bRequest,w
    Bz __USB_StdFeatureReqHandler
    Movlw cSET_FEATURE
    Subwf USB_aSetupPkt + cUSB_bRequest,w
    Bz __USB_StdFeatureReqHandler
    Movlw cGET_INTF
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest8

    Movlw (USB_aAlternateInterface & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_aAlternateInterface >> 8) & 0xFF)
    Movwf USB_wSrcH

    Movf USB_aSetupPkt + cUSB_bIntfID,w
    Addwf USB_wSrc,f
    Bsf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate RAM
    Movlw 1
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckStdRequest8
    Movlw cSET_INTF
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckStdRequest9
    Bsf USB_bInFo,cUSB_Busy
    Lfsr 2,USB_aAlternateInterface
    Movf USB_aSetupPkt + cUSB_bIntfID,w
    Movff USB_aSetupPkt + cUSB_bAltID,PLUSW2
    Return

__USB_CheckStdRequest9
    Movlw cSET_DSC
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Return
    Bra __USB_CBStdSetDscHandler

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_StdFeatureReqHandler
' Input:            None
' Output:           None
' Overview:         This routine handles the standard SET & CLEAR FEATURES requests
' Notes:            None
'
__USB_StdFeatureReqHandler
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cDEVICE_REMOTE_WAKEUP                 ' If Feature = DEVICE_REMOTE_WAKEUP
    Cpfseq USB_aSetupPkt + cUSB_bFeature
    Bra __USB_StdFeatureReqHandler1
    Movf USB_aSetupPkt,w                        ' Recipient = RCPT_DEV?
    Andlw 0x1F                                  ' Mask to lower 5 bits
    Sublw cRCPT_DEV
    Bnz __USB_StdFeatureReqHandler1
    Bsf USB_bInFo,cUSB_Busy
    Setf USB_bRemoteWakeup                      ' Preset USB_bRemoteWakeup to 1
    Movlw cSET_FEATURE                          ' Request = SET_FEATURE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Clrf USB_bRemoteWakeup

__USB_StdFeatureReqHandler1
    Movlw cENDPOINT_HALT                        ' If Feature = ENDPOINT_HALT &
    Cpfseq USB_aSetupPkt + cUSB_bFeature
__USB_StdFtrReqHandlerExit
    Return
    Movf USB_aSetupPkt,w                        ' Recepient = RCPT_EP
    Andlw 0x1F                                  ' Mask to lower 5 bits
    Sublw cRCPT_EP
    Bnz __USB_StdFtrReqHandlerExit
    Movf USB_aSetupPkt + cUSB_bEPID,w           ' USB_EPNum != 0
    Andlw 0x0F                                  ' Mask to USB_EPNum
    Bz __USB_StdFtrReqHandlerExit
    Bsf USB_bInFo,cUSB_Busy
'
' Must do address calculation here
'
    Rcall __USB_CalcEPAddress                   ' Place EndPoint buffer address in FSR2L\H
    Movlw cSET_FEATURE                          ' Request = SET_FEATURE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_StdFeatureReqHandler2
    Movlw _cUSIE | _cBSTALL
    Movwf INDF2                                 ' Put in EndPoint buffer
    Return

__USB_StdFeatureReqHandler2
    Lfsr 1,UEP0                                 ' FSR1L\H = beginning of EndPoint table
    Movf USB_aSetupPkt + cUSB_bEPID,w           ' Add in EndPoint number
    Addwf FSR1L,f
    Movlw ~cUEP_STALL
    Andwf INDF1,f
    Movlw _cUCPU | _cDAT1                       ' IN
    Btfss USB_aSetupPkt + cUSB_bEPID,cUSB_EPDir ' EPDir = 1 (IN)?
    Movlw _cUSIE | _cDAT1 | _cDTSEN             ' - OUT
    Movwf INDF2                                 ' Put in EndPoint buffer
    Return
'
' Put EndPoint buffer address in FSR2 (USB_dEp0Bo+(cUSB_EPNum*8)+(cUSB_EPDir*4))
'
__USB_CalcEPAddress
    Lfsr 2,USB_dEp0Bo                           ' Point FSR2L\H to beginning of buffer area
    Rlncf USB_aSetupPkt + cUSB_bEPID,w          ' Move EndPoint direction to Carry
    Rlcf USB_aSetupPkt + cUSB_bEPID,w           ' EndPoint number * 8
    Rlncf WREG,f
    Rlncf WREG,f
    Addwf FSR2L,f                               ' Add to FSR2
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_StdGetDscHandler
' Input:            None
' Output:           None
' Overview:         This routine handles the standard GET_DESCRIPTOR request.
' Notes:            None
'
__USB_StdGetDscHandler
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
    Movlw 0x80
    Cpfseq USB_aSetupPkt + cUSB_bmRequestType
    Return
    Movlw cUSB_INPIPES_ROM | cUSB_INPIPES_BUSY | cUSB_INPIPES_INCLUDE_ZERO
    Movwf USB_bInFo
    Movlw cUSB_DESCRIPTOR_DEVICE
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_StdGetDscHandler1
    Movlw (Desc_DeviceDescriptor & 0xFF)
    Movwf USB_wSrc
    Movlw ((Desc_DeviceDescriptor >> 8) & 0xFF)
    Movwf USB_wSrcH
'
' Set wCount
'
    Movff USB_wSrc,TBLPTRL                          ' Set source address
    Movff USB_wSrcH,TBLPTRH
    Clrf TBLPTRU
    Tblrd*                                          ' Read count
    Movff TABLAT,USB_wInCount
    Clrf USB_wInCountH
    Return

__USB_StdGetDscHandler1
    Movlw cUSB_DESCRIPTOR_CONFIGURATION
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_StdGetDscHandler2
    Clrf TBLPTRU
    Clrf TBLPTRH
    Rlncf USB_aSetupPkt + cUSB_bDscIndex,w          ' Index * 2
    Rlncf WREG,f                                    ' Index * 4
    Addlw (Desc_USB_CD_Ptr & 0xFF)                  ' Add element offset to low address
    Movwf TBLPTRL
    Movlw ((Desc_USB_CD_Ptr >> 8) & 0xFF)
    Addwfc TBLPTRH,f
    Tblrd*+                                         ' Read code memory and increment address
    Movff TABLAT,USB_wSrc                           ' Get low source address
    Tblrd*+
    Movff TABLAT,USB_wSrcH                          ' Get high source address
    Tblrd*
    Movff USB_wSrc,TBLPTRL                          ' Set amount to total length
    Movff USB_wSrcH,TBLPTRH
    Tblrd*+                                         ' Skip 2 locations
    Tblrd*+

    Tblrd*+                                         ' Skip 2 locations
    Tblrd*+

    Tblrd*+                                         ' Read count low
    Movff TABLAT,USB_wInCount
    Tblrd*+                                         ' Dummy read
    Tblrd*+                                         ' Read count high
    Movff TABLAT,USB_wInCountH
    Return

__USB_StdGetDscHandler2
    Movlw cUSB_DESCRIPTOR_STRING
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_StdGetDscHandler3
    Clrf TBLPTRU
    Clrf TBLPTRH
    Rlncf USB_aSetupPkt + cUSB_bDscIndex,w          ' Index * 2
    Rlncf WREG,f                                    ' Index * 4
    Addlw (Desc_USB_SD_Ptr & 0xFF)                  ' Add element offset to low address
    Movwf TBLPTRL
    Movlw ((Desc_USB_SD_Ptr >> 8) & 0xFF)
    Addwfc TBLPTRH,f
    Tblrd*+                                         ' Read code memory and increment address
    Movff TABLAT,USB_wSrc                           ' Get low source address
    Tblrd*+
    Movff TABLAT,USB_wSrcH                          ' Get high source address
    Tblrd*
'
' Set wCount
'
    Movff USB_wSrc,TBLPTRL                          ' Set source address
    Movff USB_wSrcH,TBLPTRH
    Tblrd*                                          ' Read count
    Movff TABLAT,USB_wInCount
    Clrf USB_wInCountH
    Return

__USB_StdGetDscHandler3
    Clrf USB_bInFo
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_StdGetStatusHandler
' Input:            None
' Output:           None
' Overview:         This routine handles the standard cGET_STATUS request
' Notes:            None
'
__USB_StdGetStatusHandler
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
    Clrf USB_aCtrlTrfData#0                         ' Initialise content
    Clrf USB_aCtrlTrfData#1
    Movf USB_aSetupPkt,w                            ' Recipient = RCPT_DEV?
    Andlw 0x1F                                      ' Mask to lower 5 bits
    Sublw cRCPT_DEV
    Bnz __USB_StdGetStatusHandler1
    Bsf USB_bInFo,cUSB_Busy
Endasm-
'
' [0]: bit0: Self-Powered Status [0] Bus-Powered [1] Self-Powered
'      bit1: USB_bRemoteWakeup   [0] Disabled    [1] Enabled
'
' If(self_power == 1)
'
$ifdef cUSE_SELF_POWER
    Bsf USB_aCtrlTrfData,0
$endif

Asm-
    Tstfsz USB_bRemoteWakeup
    Bsf USB_aCtrlTrfData,1
    Bra __USB_StdGetStatHandlerExit

__USB_StdGetStatusHandler1
    Movf USB_aSetupPkt,w                            ' Recipient = RCPT_INTF?
    Andlw 0x1F                                      ' Mask to lower 5 bits
    Sublw cRCPT_INTF
    Bnz __USB_StdGetStatusHandler2
    Bsf USB_bInFo,cUSB_Busy
    Bra __USB_StdGetStatHandlerExit

__USB_StdGetStatusHandler2
    Movf USB_aSetupPkt,w                            ' Recipient = RCPT_EP?
    Andlw 0x1F                                      ' Mask to lower 5 bits
    Sublw cRCPT_EP
    Bnz __USB_StdGetStatHandlerExit
    Bsf USB_bInFo,cUSB_Busy
'
' [0]: bit0: Halt Status [0] Not Halted [1] Halted
'
    Rcall __USB_CalcEPAddress                       ' Place EndPoint buffer address in FSR2L\H
    Movf INDF2,w
    Andlw _cBSTALL
    Bz __USB_StdGetStatHandlerExit
    Bsf USB_aCtrlTrfData,0
__USB_StdGetStatHandlerExit
    Btfss USB_bInFo,cUSB_Busy
    Return

    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_wSrcH

    Bsf USB_bInFo,cUSB_CtrlTrfMem                   ' Indicate RAM
    Movlw 2                                         ' Set count
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlEPServiceComplete
' Input:            None
' Output:           None

' Overview:         This routine wrap up the ramaining tasks in servicing a Setup Request.
'                   Its main task is to set the EndPoint controls appropriately for a given situation.
'                   See code below.
'                   There are three main scenarios:
'                       a) There was no handler for the Request,in this case a STALL should be sent out.
'                       b) The host has requested a read control transfer, endpoints are required to be setup in a specific way.
'                       c) The host has requested a write control transfer,or a control data stage is not required,endpoints are required to be setup in a specific way.
'
' Notes:            Packet processing is resumed by clearing PKTDIS bit.
'
__USB_CtrlEPServiceComplete
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
'
' PKTDIS bit is set when a Setup Transaction is received.
' Clear to resume packet processing.
'
    Bcf UCON,PKTDIS
    Btfsc USB_bInFo,cUSB_Busy
    Bra __USB_CtrlEPServiceComplete1
    Btfss USB_bOutFo,cUSB_Busy
    Bra __USB_CtrlEPServiceComplete0
    Movlw cCTRL_TRF_RX
    Movwf USB_bControlTransferState
'
' Control Write:
' <SETUP[0]><OUT[1]><OUT[0]>...<IN[1]> | <SETUP[0]>
'
' 1. Prepare IN EP to respond to early termination
'
    Clrf USB_dEp0Bi + cUSB_CNT
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Movwf USB_dEp0Bi + cUSB_STAT
'
' 2. Prepare OUT EP to receive data.
'
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Movwf USB_dEp0Bo + cUSB_STAT
    Return

__USB_CtrlEPServiceComplete0
'
' If no one knows how to service this request then stall.
' Must also prepare EP0 to receive the next SETUP transaction.
'
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Movlw _cUSIE | _cDAT0 | _cDTSEN | _cBSTALL
    Movwf USB_dEp0Bo + cUSB_STAT
    Movlw _cUSIE | _cBSTALL
    Movwf USB_dEp0Bi + cUSB_STAT
    Return

__USB_CtrlEPServiceComplete1
    Btfsc USB_bOutFo,cUSB_Busy
    Bra __USB_CtrlEPServiceComplete0
    Btfss USB_aSetupPkt,cUSB_DataDir
    Bra __USB_CtrlEPServiceComplete2
    Movf USB_wInCountH,w                            ' Make sure count does not exceed max length requested by Host
    Subwf USB_aSetupPkt + cUSB_wLength + 1,w
    Bnc _USB_CtrlEPServiceCompleteCpy
    Bnz USBCtrlEPServiceComplete11
    Movf USB_wInCount,w
    Subwf USB_aSetupPkt + cUSB_wLength,w
    Bc USBCtrlEPServiceComplete11
_USB_CtrlEPServiceCompleteCpy
    Movff USB_aSetupPkt + cUSB_wLength,USB_wInCount ' Set count to maximum
    Movff USB_aSetupPkt + cUSB_wLength + 1,USB_wInCountH
USBCtrlEPServiceComplete11
    Rcall __USB_CtrlTrfTxService                    ' Copy the data to EP0 IN buffer
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
    Movlw cCTRL_TRF_TX
    Movwf USB_bControlTransferState
'
' Control Read:
' <SETUP[0]><IN[1]><IN[0]>...<OUT[1]> | <SETUP[0]>
' 1. Prepare OUT EP to respond to early termination
'
' Note:
' If something went wrong during the control transfer, the last status stage may not be sent by the host.
' When this happens,two different things could happen depending on the host.
'   a) The host could send out a RESET.
'   b) The host could send out a new SETUP transaction without sending a RESET first.
' To properly handle case (b),the OUT EP must be setup to receive either a zero length OUT transaction,or a new SETUP transaction.
'
' Furthermore, the CNT byte should be set to prepare for
' The SETUP data (8-byte or more),and the buffer address should be pointed to USB_aSetupPkt.
'
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Movlw _cUSIE
    Movwf USB_dEp0Bo + cUSB_STAT
'
' 2. Prepare IN EP to transfer data,CNT should have been Initialised by responsible request owner.
'
    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_dEp0Bi + cUSB_ADRL
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_dEp0Bi + cUSB_ADRH
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Movwf USB_dEp0Bi + cUSB_STAT
    Return

__USB_CtrlEPServiceComplete2
    Movlw cCTRL_TRF_RX
    Movwf USB_bControlTransferState
'
' Control Write:
' <SETUP[0]><OUT[1]><OUT[0]>...<IN[1]> | <SETUP[0]>
'
' 1. Prepare IN EP to respond to early termination
'
    Clrf USB_dEp0Bi + cUSB_CNT
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Movwf USB_dEp0Bi + cUSB_STAT
'
' 2. Prepare OUT EP to receive data.
'
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Movlw _cUSIE | _cDAT1 | _cDTSEN
    Movwf USB_dEp0Bo + cUSB_STAT
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlTrfTxService
' PreCondition:     USB_wSrc, wCount and usb_stat.cUSB_CtrlTrfMem are setup properly.
' Input:            None
' Output:           None
' Overview:         This routine should be called from only two places.
'                   One from __USB_CtrlEPServiceComplete() and one from __USB_CtrlTrfInHandler(). It takes care of managing a transfer over multiple USB transactions.
' Notes:            This routine works with isochronous EndPoint larger than 256 bytes and is shown here as an example of how to deal with BC9 and BC8.
'                   In reality,a control EndPoint can never be larger than 64 bytes.
'
__USB_CtrlTrfTxService
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movf USB_wInCount,w                         ' Preset Count bytes to send
    Movwf USB_wTemp
    Movf USB_wInCountH,w
    Movwf USB_wTempH
    Sublw ((cUSB_EP0_BUFF_SIZE >> 8) & 0xFF)    ' Check for count less than maximium length
    Bnz __USB_CtrlTrfTxServiceNotEQ             ' Not equal so we don't need to check the bottom byte
    Movf USB_wInCount,w
    Sublw (cUSB_EP0_BUFF_SIZE & 0xFF)
    Bz __USB_CtrlTrfTxServiceNotLess            ' Equal to
__USB_CtrlTrfTxServiceNotEQ
    Bnc __USB_CtrlTrfTxServiceGtr               ' Greater than
    Movlw cSHORT_PKT_NOT_USED
    Cpfseq USB_bShortPacketStatus
    Bra __USB_CtrlTrfTxServiceNotEQ1
    Movlw cSHORT_PKT_PENDING
    Movwf USB_bShortPacketStatus
    Bra __USB_CtrlTrfTxServiceNotLess

__USB_CtrlTrfTxServiceNotEQ1
    Movlw cSHORT_PKT_PENDING
    Cpfseq USB_bShortPacketStatus
    Bra __USB_CtrlTrfTxServiceNotLess
    Movlw cSHORT_PKT_SENT
    Movwf USB_bShortPacketStatus
    Bra __USB_CtrlTrfTxServiceNotLess

__USB_CtrlTrfTxServiceGtr
    Movlw (cUSB_EP0_BUFF_SIZE & 0xFF)           ' Send buffer full of bytes
    Movwf USB_wTemp
    Movlw ((cUSB_EP0_BUFF_SIZE >> 8) & 0xFF)
    Movwf USB_wTempH
__USB_CtrlTrfTxServiceNotLess
    Movf USB_dEp0Bi + cUSB_STAT,w               ' Get full STAT byte
    Andlw 0xFC                                  ' Clear bottom bits
    Iorwf USB_wTempH,w                          ' Put in high bits of bytes to send
    Movwf USB_dEp0Bi + cUSB_STAT                ' Save it out
    Movf USB_wTemp,w                            ' Get bytes to be sent from the buffer for subtract
    Movwf USB_dEp0Bi + cUSB_CNT                 ' Save low number of bytes to send while we're here
    Subwf USB_wInCount,f                        ' Subtract the number of bytes just about to be sent from the total.
    Movf USB_wTempH,w
    Subwfb USB_wInCountH,f
    Lfsr 2,USB_aCtrlTrfData                     ' Set destination pointer
    Movf USB_wTempH,w                           ' Check high byte for 0
    Iorwf USB_wTemp,w                           ' Check low byte for 0
    Bz __USB_CtrlTrfTxServiceExit               ' If both 0 then nothing to send this time
    Btfsc USB_bInFo,cUSB_CtrlTrfMem             ' ROM or RAM?
    Bra __USB_CtrlTrfTxServiceRam               ' RAM
    Movff USB_wSrc,TBLPTRL                      ' Move source pointer to TBLPTR
    Movff USB_wSrcH,TBLPTRH
    Clrf TBLPTRU

__USB_CtrlTrfTxServiceRomLoop
    Tblrd*+
    Movff TABLAT,POSTINC2                       ' Copy one buffer to the other
    Tblrd*+                                     ' Skip high location
    Decf USB_wTemp,f                            ' Count down number of bytes
    Bnz __USB_CtrlTrfTxServiceRomLoop
    Decf USB_wTempH,f
    Bc __USB_CtrlTrfTxServiceRomLoop
    Movff TBLPTRL,USB_wSrc                      ' Update source pointer
    Movff TBLPTRH,USB_wSrcH
    Return

__USB_CtrlTrfTxServiceRam
    Movff USB_wSrc,FSR1L                        ' \ Move source pointer to FSR1L\H
    Movff USB_wSrcH,FSR1H                       ' /
__USB_CtrlTrfTxServiceRamLoop
    Movff POSTINC1,POSTINC2                     ' Copy one buffer to the other
    Decf USB_wTemp,f                            ' Count down number of bytes
    Bnz __USB_CtrlTrfTxServiceRamLoop
    Decf USB_wTempH,f
    Bc __USB_CtrlTrfTxServiceRamLoop
    Movff FSR1L,USB_wSrc                        ' Update source pointer
    Movff FSR1H,USB_wSrcH
__USB_CtrlTrfTxServiceExit
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CtrlTrfRxService
' PreCondition:     USB_wDst and wCount are setup properly.
'                   USB_wSrc is always &USB_aCtrlTrfData
'                   usb_stat.cUSB_CtrlTrfMem is always _RAM.
'                   wCount should be set to 0 at the start of each control transfer.
' Input:            None
' Output:           None
' Notes:            None
'
__USB_CtrlTrfRxService
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
    Movf USB_dEp0Bo + cUSB_CNT,w
    Movwf USB_wTemp
'
' Accumulate total number of bytes read
'
    Tstfsz USB_wOutCountH                           ' If high byte isn't 0 then not less than
    Bra __USB_CtrlTrfRxService1
    Movf USB_wOutCount,w                            ' Get low byte
    Cpfslt USB_wTemp
    Movwf USB_wTemp
__USB_CtrlTrfRxService1
    Movf USB_wTemp,w
    Subwf USB_wOutCount,f
    Btfss STATUS,C
    Decf USB_wOutCountH,f
    Lfsr 1,USB_aCtrlTrfData                         ' Point FSR1L\H to source
    Movff USB_wDst,FSR2L                            ' Move destination pointer to FSR2
    Movff USB_wDstH,FSR2H
    Movf USB_wTemp,w                                ' Check low byte for 0
    Bz __USB_CtrlTrfRxServiceExit                   ' If both 0 then nothing to send this time
__USB_CtrlTrfRxServiceLoop
    Movff POSTINC1,POSTINC2                         ' Copy one buffer to the other
    Decf USB_wTemp,f                                ' Count down number of bytes
    Bnz __USB_CtrlTrfRxServiceLoop
    Movff FSR2L,USB_wDst                            ' Update destination pointer
    Movff FSR2H,USB_wDstH
__USB_CtrlTrfRxServiceExit
    Movf USB_wOutCountH,w                           ' Check high byte for 0
    Iorwf USB_wOutCount,w                           ' Check low byte for 0
    Bz __USB_CtrlTrfRxService2
'
' Don't have to worry about overwriting _KEEP bit because if _KEEP was set, TRNIF would not have been generated in the first place.
'
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
    Movlw cUSB_EP0_BUFF_SIZE
    Movwf USB_dEp0Bo + cUSB_CNT
    Movlw (USB_aCtrlTrfData & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aCtrlTrfData >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Btg USB_dEp0Bo + cUSB_STAT,cUSB_DTS
    Return

__USB_CtrlTrfRxService2
    Movlw (USB_aSetupPkt & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRL
    Movlw ((USB_aSetupPkt >> 8) & 0xFF)
    Movwf USB_dEp0Bo + cUSB_ADRH
    Bcf USB_bOutFo,cUSB_Busy
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_StdSetCfgHandler
' Input:            None
' Output:           None
' Overview:         This routine first disables all endpoints by clearing UEP registers.
'                   It then configures (Initialises) endpoints by calling the callback subroutine USBCBInitDevClass().
' Notes:            None
'
__USB_StdSetCfgHandler
    Movlb (_usb_ram >> 8)                           ' Point to the USB RAM bank
'
' This will generate a zero length packet inPipes[0].USB_bInFo.bits.cUSB_Busy = 1'
'
    Bsf USB_bInFo,cUSB_Busy
'
' Disable all endpoints except EndPoint 0
'
    Lfsr 2,UEP1                                     ' Reset all non-EP0 UEPn registers
    Movlw cUSB_MAX_EP_NUMBER
__USB_StdSetCfgHandlerClrEPLoop
    Clrf POSTINC2
    Decfsz WREG,f
    Bra __USB_StdSetCfgHandlerClrEPLoop
'
' Clear the alternate interface settings
'
    Lfsr 2,USB_aAlternateInterface                  ' Clear USB_aAlternateInterface array
    Movlw cUSB_MAX_NUM_INT
__USB_StdSetCfgHandlerClrAltLoop
    Clrf POSTINC2
    Decfsz WREG,f
    Bra __USB_StdSetCfgHandlerClrAltLoop
'
' Set the current configuration
'
    Movf USB_aSetupPkt + cUSB_bConfigValue,w
    Movwf USB_bActiveConfiguration
    Bnz __USB_StdSetCfgHandler1
    Movlw cADDRESS_STATE
    Movwf USB_bDeviceState
    Return

__USB_StdSetCfgHandler1
    Movlw cCONFIGURED_STATE
    Movwf USB_bDeviceState
    Bra __USB_CBInitEP
Endasm-

'--------------------------------------------------------------------------------------------------------------------
' Enable and configure OUT EndPoint
' Input     : GEN - EndPoint number
'           : FSR1L\H - EndPoint buffer address USB_wTemp is options
'           : WREG - EndPoint buffer size
' Output    : None
' Notes     : None
'
__USB_EnableEndpointOut:
    POSTINC1 = WREG                         ' Save maximum count at beginning of EndPoint buffer and increment address
    Lfsr 2,UEP0                             ' Point FSR2L\H to the beginning of the EndPoint table
    FSR2L = FSR2L + GEN                     ' Add the EndPoint number
    Movff USB_wTemp,INDF2                   ' Enable In EndPoint
'
' Point FSR0L\H to EndPoint BDT
'
    Rlncf GEN,w                             ' EndPoint number * 8
    Rlncf WREG,f
    Rlncf WREG,f
    Lfsr 0,USB_dEp0Bo                       ' Point FSR0 to beginning of BDT area
    FSR0L = FSR0L + WREG                    ' Add EndPoint offset to FSR0
'
' Set EndPoint buffer address from FSR1L\H
'
    WREG = cUSB_ADRL                        ' \
    PLUSW0 = FSR1L                          ' |
    WREG = cUSB_ADRH                        ' | Point to ADRL\H
    PLUSW0 = FSR1H                          ' /
'
' Set CNT to maximum amount
'
    WREG = POSTDEC1                         ' Move back EndPoint buffer pointer
    Inc FSR0L                               ' Point to CNT
    POSTDEC0 = INDF1                        ' Set maximum count and point back to STAT
'
' Set EndPoint status
'
    Clear INDF0.cUSB_UOWN
    Set INDF0.cUSB_DTS
    Return
'--------------------------------------------------------------------------------------------------------------------
' Enable and configure IN EndPoint
' Input     : GEN - EndPoint number
'           : FSR1L\H - EndPoint buffer address USB_wTemp is options
'           : WREG - EndPoint buffer size
' Output    : None
' Notes     : None
'
__USB_EnableEndpointIn:
    POSTINC1 = WREG                         ' Save maximum count at beginning of EndPoint buffer and increment address
    Lfsr 2,UEP0                             ' Point FSR2L\H to the beginning of the EndPoint table
    FSR2L = FSR2L + GEN                     ' Add the EndPoint number
    Movff USB_wTemp,INDF2                   ' Enable In EndPoint
'
' Point FSR0L\H to EndPoint BDT
'
    Rlncf GEN,w                             ' EndPoint number * 8
    Rlncf WREG,f
    Rlncf WREG,f
    Lfsr 0,USB_dEp0Bi                       ' Point FSR0 to beginning of BDT area
    FSR0L = FSR0L + WREG                    ' Add EndPoint offset to FSR0L\H
'
' Set EndPoint buffer address from FSR1L\H
'
    WREG = cUSB_ADRL                        ' \
    PLUSW0 = FSR1L                          ' |
    WREG = cUSB_ADRH                        ' | Point to cUSB_ADRL\H
    PLUSW0 = FSR1H                          ' /
'
' Set EndPoint status
'
    Clear INDF0.cUSB_UOWN
    Set INDF0.cUSB_DTS
    Return

'-------------------------------------------------------------------------------------------------
' Poll the USB subroutines in order to keep the USB interface attached
' Input     : None
' Output    : USB_bStatus used to store the status variable USB_bDeviceState:
'               DETACHED_STATE    = $00
'               ATTACHED_STATE    = $01
'               POWERED_STATE     = $02
'               DEFAULT_STATE     = $04
'               ADR_PENDING_STATE = $08
'               ADDRESS_STATE     = $10
'               CONFIGURED_STATE  = $20
' Notes     : Polling must be done at least every 10ms
'           : PP1\H used to save\restore the FSR2L\H registers if a stack is implemented in the BASIC code
'           : Alters the value of WREG
'
#ifdef __USB_SERVICE_REQ
__USB_Poll_:
#ifdef WATCHDOG_REQ
    Clrwdt
#endif
#ifSym __STACK_SIZE                         ' Are we using a software stack?
    USB_wSaveSFR = USB_wFSR2                ' Yes. So save FSR2L\H
#endIfSym
    Call __USB_DeviceTasks                  ' Keep the USB alive
    USB_bStatus = USB_bDeviceState          ' Copy the status into USB_STATUS
    Movlb 0                                 ' Point back to RAM bank 0
    Bra __USB_Exit                          ' Exit via __USB_Exit restoring FSR2L\H if required
#endif ' __USB_SERVICE_REQ

'--------------------------------------------------------------------------------------------------------------------
' Get from OUT EndPoint for RX with auto length detection via a null terminator
' Input     : GEN - The EndPoint the data will be transmitted on
'           : FSR1 - Address of destination buffer
'           : WREG - Length of the data needing to be sent
' Output    : FSR1 - Updated destination buffer pointer
'           : WREG - Amount of bytes received
'           : Carry - Set for buffer not available
' Notes     : None
'
#ifdef __USBIN_AUTO_REQ                     ' Are we using USBin with auto length detect?
__USB_GetBuffer_Auto_:                      ' Yes. So... Include this block of code
    PRODL = WREG                            ' Save the maximum amount of bytes to receive
    Rcall __USB_CheckForConfigured          ' Check to see if we're configured
    Bnz __USB_Exit                          ' Exit if not configured
    Rcall __USB_PointToOutEndPoint          ' Point FSR0L\H to the requested EndPoint In BDT
    Bc __USB_Exit                           ' Exit if we don't own the bus
    Rcall __USB_RX_GetEndPointBuffAddress   ' Get the EndPoint buffer address to FSR2L\H
    Bz __USB_GetPutUSB                      ' Nothing received, so exit via __USB_GetPutUSB
'
' Copy the EndPoint buffer to the destination buffer with null test
'
    WREG = POSTINC2                         ' Get a byte from the buffer to WREG
    POSTINC1 = WREG                         ' Transfer it to the user's buffer
    Bz $ + 6                                ' Exit if null found
    Decfsz PRODL,f                          ' Count amount of bytes received
    Bra $ - 6
    Bra __USB_GetPutUSB                     ' Exit via __USB_GetPutUSB
#endif  ' __USBIN_REQ

'--------------------------------------------------------------------------------------------------------------------
' Get from OUT EndPoint for RX
' Input     : GEN - The EndPoint the data will be transmitted on
'           : FSR1 - Address of destination buffer
'           : WREG - Length of the data needing to be sent
' Output    : FSR1 - Updated destination buffer pointer
'           : WREG - Amount of bytes received
'           : Carry - Set for buffer not available
' Notes     : None
'
#ifdef __USBIN_REQ                          ' Are we using USBin?
__USB_GetBuffer_:                           ' Yes. So... Include this block of code
    PRODL = WREG                            ' Save the amount of bytes to receive
    Rcall __USB_CheckForConfigured          ' Check to see if we're configured
    Bnz __USB_Exit                          ' Exit if not configured
    Rcall __USB_PointToOutEndPoint          ' Point FSR0L\H to the requested EndPoint In BDT
    Bc __USB_Exit                           ' Exit if we don't own the bus
    Rcall __USB_RX_GetEndPointBuffAddress   ' Get the EndPoint buffer address to FSR2L\H
    Bz __USB_GetPutUSB                      ' Nothing received, so exit via __USB_GetPutUSB
'
' Copy the EndPoint buffer to the destination buffer
'
    POSTINC1 = POSTINC2                     ' Copy EndPoint buffer to destination buffer
    Decfsz WREG,f                           ' Count amount of bytes received
    Bra $ - 6
    Bra __USB_GetPutUSB                     ' Exit via __USB_GetPutUSB
#endif  ' __USBIN_REQ

'---------------------------------------------------------------------------------------------------------------
' Fill IN EndPoint for TX from code memory
' Input     : GEN - EndPoint number
'           : TBLPTRL\H - Code memory source buffer pointer
'           : WREG - The amount of bytes to transmit
' Output    : WREG - Amount sent
'           : Carry - Set if buffer not available
' Notes     : None
'
#ifdef __USBOUT_CODE_AUTO_REQ               ' Do we require transmit from code memory with auto length?
__USB_PutCode_Auto_:                        ' Yes. So...
'
' Find the length of the code memory TX buffer by looking for a null terminator and place the length into WREG
'
    USB_wSaveSFR = USB_wTBLPTR              ' Save the contents of TBLPTRL\H because "__CodeMemory__Length_" alters it
    Call __CodeMemory__Length_              ' Return the length of the code memory string in WREG
    USB_wTBLPTR = USB_wSaveSFR              ' Restore the contents of TBLPTRL\H
#endif  ' __USBOUT_CODE_AUTO_REQ

#ifdef __USBOUT_CODE_REQ                    ' Do we require transmit from the code memory buffer?
__USB_PutCode_:                             ' Yes. So...
    PRODL = WREG                            ' Save the amount of bytes to transmit
    Rcall __USB_CheckForConfigured          ' Check to see if we're configured
    Bnz __USB_Exit                          ' Exit if not configured
    Rcall __USB_PointToInEndPoint           ' Point FSR0L\H to the requested EndPoint In BDT
    Bc __USB_Exit                           ' Exit if we don't own the bus
    Rcall __USB_TX_GetEndPointBuffAddress   ' Get the EndPoint buffer address to FSR2L\H
    Bz __USB_GetPutUSB                      ' Zero bytes to send so exit via __USB_GetPutUSB
'
' Copy the code memory source buffer to the EndPoint buffer
'
    Tblrd*+                                 ' Read a byte from code memory
    Movff TABLAT,POSTINC2                   ' Copy the code memory source buffer to the EndPoint buffer
    Decfsz WREG,f                           ' Count the amount of bytes to transfer
    Bra $ - 8
    Bra __USB_GetPutUSB                     ' Exit via __USB_GetPutUSB
#endif  ' __USBOUT_CODE_REQ

'---------------------------------------------------------------------------------------------------------------
' Fill IN EndPoint for TX from RAM
' Input     : GEN - EndPoint number
'           : FSR1 - Source buffer pointer
'           : WREG - The amount of bytes to transmit
' Output    : FSR1 - Updated source buffer pointer
'           : WREG - Amount sent
'           : Carry - Set if buffer not available
' Notes     : None
'
#ifdef __USBOUT_RAM_AUTO_REQ                ' Do we require transmit from RAM with auto length?
__USB_PutRAMBuffer_Auto_:                   ' Yes. So...
'
' Find the length of the TX buffer by looking for a null terminator and place the length into WREG
'
    USB_wFSR0 = USB_wFSR1                   ' Transfer the tx buffer's address to FSR0L\H
    WREG = 0                                ' Reset the count variable
    Movf POSTINC0,f                         ' Test for null and auto increment to next source address
    Bz __USB_PutRAMBuffer_                  ' Exit when null found
    Incfsz WREG,f                           ' Increment the counting variable and prevent an infinite loop
    Bra $ - 6                               ' Keep looking until null found or a count of 255 reached
#endif  ' __USBOUT_RAM_AUTO_REQ

#ifdef __USBOUT_RAM_REQ                     ' Do we require transmit from the RAM buffer?
__USB_PutRAMBuffer_:                        ' Yes. So...
    PRODL = WREG                            ' Save the amount of bytes to transmit
    Rcall __USB_CheckForConfigured          ' Check to see if we're configured
    Bnz __USB_Exit                          ' Exit if not configured
    Rcall __USB_PointToInEndPoint           ' Point FSR0L\H to the requested EndPoint in BDT
    Bc __USB_Exit                           ' Exit if we don't own the bus
    Rcall __USB_TX_GetEndPointBuffAddress   ' Get the EndPoint buffer address to FSR2L\H
    Bz __USB_GetPutUSB                      ' Zero bytes to send so exit via __USB_GetPutUSB
'
' Copy the source buffer to the EndPoint buffer
'
    POSTINC2 = POSTINC1                     ' Copy the source buffer to the EndPoint buffer
    Decfsz WREG,f                           ' Count the amount of bytes to transfer
    Bra $ - 6
    Bra __USB_GetPutUSB                     ' Exit via __USB_GetPutUSB
#endif  ' __USBOUT_RAM_REQ

'---------------------------------------------------------------
' Check to see if we're configured
'
' Input     : None
' Output    : Zero flag (STATUS.2) is Clear if not configured
' Notes     : None
'
#ifdef __USBINOUT_REQ
__USB_CheckForConfigured:
#ifSym __STACK_SIZE                         ' Are we using a software stack?
    USB_wSaveSFR = USB_wFSR2                ' Yes. So save FSR2L\H
#endIfSym
    Movlb (_usb_ram >> 8)                   ' Point to the USB RAM bank
    Call __USB_Poll_
    WREG = cCONFIGURED_STATE                ' \ We might not be configured yet
    Subwf USB_bDeviceState,w                ' /
    Set STATUS.0                            ' Set Carry for possible error return
    WREG = 0                                ' 0 bytes sent so far
    Movlb 0                                 ' Point back to RAM bank 0
    Return

'------------------------------------------------------------------
' Transmit the contents of the buffer
'
' Input     : None
' Output    : None
' Notes     : None
'
__USB_GetPutUSB:
    WREG = _cDTSMASK                        ' Save only the DTS bit
    INDF0 = INDF0 & WREG
    Toggle INDF0.cUSB_DTS                   ' Toggle the DTS bit
    WREG = _cUSIE | _cDTSEN                 ' Turn ownership to SIE
    INDF0 = INDF0 | WREG
    WREG = PRODL                            ' Return amount of bytes received
    Clear STATUS.0                          ' Clear the Carry flag for an OK return
__USB_Exit:
#ifSym __STACK_SIZE                         ' Are we using a software stack?
    USB_wFSR2 = USB_wSaveSFR                ' Yes. So restore FSR2L\H
#endIfSym
    Return
#endif ' __USBINOUT_REQ
'---------------------------------------------------------------
' Point FSR0L\H to requested EndPoint In BDT
'
' Input     : GEN holds the EndPoint number
' Output    : Carry flag (STATUS.0) is Set if we don't own the bus
' Notes     : None
'
#ifdef __USBOUT_REQ
__USB_PointToInEndPoint:
    Rlncf GEN,w                             ' \
    Rlncf WREG,f                            ' | EndPoint number * 8
    Rlncf WREG,f                            ' /
    Lfsr 0,USB_dEp0Bi                       ' Point FSR0 to beginning of BDT area
    FSR0L = FSR0L + WREG                    ' Add EndPoint offset to FSR0
    WREG = 0                                ' 0 bytes sent, so far
    Clear STATUS.0                          ' Default to an OK return
    Btfsc INDF0,cUSB_UOWN                   ' Who owns the buffer (STAT,UOWN)?
    Set STATUS.0                            ' Set the carry to indicate we don't own the bus
    Return

'------------------------------------------------------------------
' Get Transmission EndPoint buffer address to FSR2L\H
'
' Input     : None
' Output    : WREG holds the amount of bytes to transmit
'           : Zero flag (STATUS.2) is set if no bytes to send
' Notes     : None
'
__USB_TX_GetEndPointBuffAddress:
    WREG = cUSB_ADRL                        ' \
    FSR2L = PLUSW0                          ' |
    WREG = cUSB_ADRH                        ' | Point FSR2L\H to ADRL\H
    FSR2H = PLUSW0                          ' /
    WREG = 255                              ' \
    WREG = PLUSW2                           ' / Get maximum length from start of EndPoint buffer
    Cpfslt PRODL                            ' Find the amount of bytes to send this time
    PRODL = WREG                            ' Too many, so send the maximum allowed length
    Inc FSR0L                               ' Point to CNT
    WREG = PRODL                            ' Get amount to send
    POSTDEC0 = WREG                         ' Place the length into CNT and point back to STAT
    Return
#endif ' __USBOUT_REQ
'---------------------------------------------------------------
' Point FSR0L\H to requested EndPoint Out BDT
'
' Input     : GEN holds the EndPoint number
' Output    : Carry flag (STATUS.0) is Set if we don't own the bus
' Notes     : None
'
#ifdef __USBIN_REQ
__USB_PointToOutEndPoint:
    Rlncf GEN,w                             ' \
    Rlncf WREG,f                            ' | EndPoint number * 8
    Rlncf WREG,f                            ' /
    Lfsr 0,USB_dEp0Bo                       ' Point FSR0 to beginning of BDT area
    FSR0L = FSR0L + WREG                    ' Add EndPoint offset to FSR0
    WREG = 0                                ' 0 bytes sent, so far
    Clear STATUS.0                          ' Default to an OK return
    Btfsc INDF0,cUSB_UOWN                   ' Who owns the buffer (STAT,UOWN)?
    Set STATUS.0                            ' Set the carry to indicate we don't own the bus
    Return

'------------------------------------------------------------------
' Get Reception EndPoint buffer address to FSR2L\H
'
' Input     : None
' Output    : WREG holds the amount of bytes received
'           : Zero flag (STATUS.2) is set if no bytes received
' Notes     : None
'
__USB_RX_GetEndPointBuffAddress:
    WREG = cUSB_ADRL                        ' \
    FSR2L = PLUSW0                          ' |
    WREG = cUSB_ADRH                        ' | Point FSR2L\H to ADRL\H
    FSR2H = PLUSW0                          ' /
    WREG = PREINC0                          ' Get CNT
    Cpfslt PRODL                            ' Make sure it's not greater than the buffer
    PRODL = WREG                            ' Save the amount of bytes to receive
    WREG = 255                              ' \
    POSTDEC0 = PLUSW2                       ' / Get maximum amount from in front of EndPoint buffer and point back to STAT
    WREG = PRODL                            ' Place the amount of bytes received into WREG
    Return
#endif ' __USBIN_REQ

$endif ' _USB_DEV_INC_
