$ifndef _USB_HID_INC_
$define _USB_HID_INC_

$define cUSB_USE_HID         ' Signal that we're using USB_HID
'
' Class-Specific Requests
'
    Symbol cGET_REPORT            = $01
    Symbol cGET_IDLE              = $02
    Symbol cGET_PROTOCOL          = $03
    Symbol cSET_REPORT            = $09
    Symbol cSET_IDLE              = $0A
    Symbol cSET_PROTOCOL          = $0B
'
' Class Descriptor Types
'
    Symbol cDSC_HID               = $21
    Symbol cDSC_RPT               = $22
    Symbol cDSC_PHY               = $23
'
' Protocol Selection
'
    Symbol cBOOT_PROTOCOL         = $00
    Symbol cRPT_PROTOCOL          = $01
'
' HID Interface Class Code
'
    Symbol cHID_INTF              = $03
'
' HID Interface Class SubClass Codes
'
    Symbol cBOOT_INTF_SUBCLASS    = $01
'
' HID Interface Class Protocol Codes
'
    Symbol cHID_PROTOCOL_NONE     = $00
    Symbol cHID_PROTOCOL_KEYBOARD = $01
    Symbol cHID_PROTOCOL_MOUSE    = $02

    Include "USB_Dev.inc"                       ' Include the main USB routines

Asm-
'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CheckHIDRequest
' Input:            None
' Output:           None
' Overview:         This routine checks the setup data packet to see if it knows how to handle it
' Notes:            None
'
__USB_CheckHIDRequest
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movf USB_aSetupPkt,w                        ' Recipient = RCPT_INTF?
    Andlw 0x1F                                  ' Mask to lower 5 bits
    Sublw cRCPT_INTF
    Bnz __USB_CheckHIDRequestExit
    Movlw cHID_INTF_ID                          ' IntfID = HID_INTF_ID?
    Cpfseq USB_aSetupPkt + cUSB_bIntfID
__USB_CheckHIDRequestExit
    Return
'
' There are two standard requests that hid.c may support.
' 1. cGET_DSC(cDSC_HID,cDSC_RPT,cDSC_PHY)'
' 2. cSET_DSC(cDSC_HID,cDSC_RPT,cDSC_PHY)'
'
    Movlw cGET_DSC                              ' Request = GET_DSC?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckHIDRequestClass
    Movlw cDSC_HID                              ' DescriptorType = DSC_HID?
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_CheckHIDRequest1
    Movlw 1                                     ' USB_bActiveConfiguration = 1?
    Cpfseq USB_bActiveConfiguration
    Bra __USB_CheckHIDRequestClass
    Movlw (Desc_HIDDescriptor1 & 0xFF)
    Movwf USB_wSrc
    Movlw ((Desc_HIDDescriptor1 >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movff USB_wSrc,TBLPTRL                      ' Set source address
    Movff USB_wSrcH,TBLPTRH
    Tblrd*                                      ' Read count
    Movff TABLAT,USB_wInCount
    Clrf USB_wInCountH
    Bcf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate ROM
    Bsf USB_bInFo,cUSB_IncludeZero              ' Include a trailing zero packet
    Bsf USB_bInFo,cUSB_Busy
    Bra __USB_CheckHIDRequestClass
__USB_CheckHIDRequest1
    Movlw cDSC_RPT                              ' DescriptorType = DSC_RPT?
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_CheckHIDRequest2
    Movlw 1                                     ' USB_bActiveConfiguration = 1?
    Cpfseq USB_bActiveConfiguration
    Bra __USB_CheckHIDRequestClass
    Movlw (Desc_ReportDescriptor1 & 0xFF)
    Movwf USB_wSrc
    Movlw ((Desc_ReportDescriptor1 >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movlw (Desc_ReportDescriptor1Len & 0xFF)    ' Set Count
    Movwf TBLPTRL
    Movlw ((Desc_ReportDescriptor1Len >> 8) & 0xFF)
    Movwf TBLPTRH
    Clrf TBLPTRU
    Tblrd*+                                     ' Read count low
    Movff TABLAT,USB_wInCount
    Tblrd*+                                     ' Skip next
    Tblrd*                                      ' Read count high
    Movff TABLAT,USB_wInCountH
    Bcf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate ROM
    Bsf USB_bInFo,cUSB_IncludeZero              ' Include a trailing zero packet
    Bsf USB_bInFo,cUSB_Busy
    Bra __USB_CheckHIDRequestClass
__USB_CheckHIDRequest2
    Movlw cDSC_PHY                              ' DescriptorType = DSC_PHY?
    Cpfseq USB_aSetupPkt + cUSB_bDescriptorType
    Bra __USB_CheckHIDRequestClass
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
__USB_CheckHIDRequestClass
    Movf USB_aSetupPkt,w                        ' RequestType = CLASS?
    Andlw 0x60                                  ' Mask to correct bits
    Sublw (cCLASS) << 5
    Bnz __USB_CheckHIDRequestExit
    Movlw cGET_REPORT                           ' Request = GET_REPORT?
    Subwf USB_aSetupPkt + cUSB_bRequest,w
    Bz __HID_GetSetReportHandler                ' Yes
    Movlw cSET_REPORT                           ' Request = SET_REPORT?
    Subwf USB_aSetupPkt + cUSB_bRequest,w
    Bz __HID_GetSetReportHandler                ' Yes
    Movlw cGET_IDLE                             ' Request = GET_IDLE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckHIDRequestClass1
    Movlw (USB_bIdleRate & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_bIdleRate >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movlw 1
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Bsf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate RAM
    Bsf USB_bInFo,cUSB_IncludeZero              ' Include a trailing zero packet
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckHIDRequestClass1
    Movlw cSET_IDLE                             ' Request = SET_IDLE?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckHIDRequestClass2
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
    Movff USB_aSetupPkt + cUSB_wValue + 1,USB_bIdleRate
    Return

__USB_CheckHIDRequestClass2
    Movlw cGET_PROTOCOL                         ' Request = GET_PROTOCOL?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Bra __USB_CheckHIDRequestClass3
    Movlw (USB_bActiveProtocol & 0xFF)
    Movwf USB_wSrc
    Movlw ((USB_bActiveProtocol >> 8) & 0xFF)
    Movwf USB_wSrcH
    Movlw 1
    Movwf USB_wInCount
    Clrf USB_wInCountH
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_CtrlTrfMem               ' Indicate RAM
    Bsf USB_bInFo,cUSB_Busy
    Return

__USB_CheckHIDRequestClass3
    Movlw cSET_PROTOCOL                         ' Request = cSET_PROTOCOL?
    Cpfseq USB_aSetupPkt + cUSB_bRequest
    Return
    Clrf USB_bInFo
    Bsf USB_bInFo,cUSB_Busy
    Movff USB_aSetupPkt + cUSB_wValue,USB_bActiveProtocol
__HID_GetSetReportHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' USB Callback Functions
'--------------------------------------------------------------------------------------------------------------------
' The USB firmware stack will call the callback functions USBCBxxx() in response to certain USB related events.
' For example,if the host PC is powering down,it will stop sending out Start of Frame (SOF) packets to your device.
' In response to this,all USB devices are supposed to decrease their power consumption from the USB Vbus to <2.5mA each.
' The USB module detects this condition (which according to the USB specifications is 3+ms of no bus activity/SOF packets) and then calls the __USB_CBSuspend() subroutine.
' You should modify these callback functions to take appropriate actions for each of these conditions.
' For example,in the __USB_CBSuspend(),you may wish to add code that will decrease power consumption from Vbus to <2.5mA (such as by clock switching,turning off LEDs,
' putting the microcontroller to sleep,etc.).
' Then,in the __USB_CBWakeFromSuspend() subroutine,you may then wish to add code that undoes the power saving things done in the __USB_CBSuspend() subroutine.

' The __USB_CBSendResume() subroutine is special,in that the USB stack will not automatically call this subroutine.
' This subroutine is meant to be called from the application firmware instead.  See the additional comments near the subroutine.

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBSuspend
' Input:            None
' Output:           None
' Overview:         Call back that is invoked when a USB suspend is detected
' Notes:            None
'
__USB_CBSuspend
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBWakeFromSuspend
' Input:            None
' Output:           None
' Overview:         The host may put USB peripheral devices in low power suspend mode (by "sending" 3+ms of idle).
'                   Once in suspend mode,the host may wake the device back up by sending non-idle state signalling.
'                   This call back is invoked when a wakeup from USB suspend is detected.
' Notes:            None
'
__USB_CBWakeFromSuspend
' If clock switching or other power savings measures were taken when executing the __USB_CBSuspend() subroutine,
' now would be a good time to switch back to normal full power run mode conditions.
' The host allows a few milliseconds of wakeup time,after which the device must be fully back to normal,
' and capable of receiving and processing USB packets.
' In order to do this,the USB module must receive correct clocking (IE: 48MHz clock must be available to SIE for full speed USB operation).
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CB_SOF_Handler
' Input:            None
' Output:           None
' Overview:         The USB host sends out a SOF packet to full-speed devices every 1 ms.
'                   This interrupt may be useful for isochronous pipes.
'                   End designers should implement callback routine as necessary.
' Notes:            None
'
__USB_CB_SOF_Handler
' No need to clear UIRbits.SOFIF to 0 here.
' Callback caller is already doing that.
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBErrorHandler
' Input:            None
' Output:           None
' Overview:         The purpose of this callback is mainly for debugging during development.
'                   Check UEIR to see which error causes the interrupt.
' Notes:            None
'
__USB_CBErrorHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBCheckOtherReq
' Input:            None
' Output:           None
' Overview:         When SETUP packets arrive from the host,some firmware must process the request and respond appropriately to fulfill the request.
'                   Some of the SETUP packets will be for standard USB "chapter 9" (as in,fulfilling chapter 9 of the official USB specifications) requests,
'                   while others may be specific to the USB device class that is being implemented.
'                   For example,a HID class device needs to be able to respond to "GET REPORT" type of requests.
'                   This is not a standard USB chapter 9 request,and therefore not handled by usb_device.c.
'                   Instead this request should be handled by class specific firmware,such as that contained in usb_function_hid.c.
' Notes:            None
'
__USB_CBCheckOtherReq
    Bra __USB_CheckHIDRequest

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBStdSetDscHandler
' Input:            None
' Output:           None
' Overview:         The __USB_CBStdSetDscHandler() callback subroutine is called when a SETUP,cUSB_bRequest: SET_DESCRIPTOR request arrives.
'                   Typically SET_DESCRIPTOR requests are not used in most applications,and it is optional to support this type of request.
' Notes:            None
'
__USB_CBStdSetDscHandler
    Return

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBInitEP
' Input:            None
' Output:           None
' Overview:         This subroutine is called when the device becomes initialised,which occurs after the host sends a SET_CONFIGURATION (cUSB_wValue not = 0) request.
'                   This callback subroutine should initialise the endpoints for the device's usage according to the current configuration.
' Notes:            None
'
__USB_CBInitEP
'
' Enable the HID EndPoint
' Enable and configure OUT EndPoint
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cHID_EP
    Movwf GEN                                   ' Endpoint number
    Movlw cUSB_OUT_ENABLED | cUSB_IN_ENABLED | cUSB_HANDSHAKE_ENABLED | cUSB_DISALLOW_SETUP
    Movwf USB_wTemp                             ' Options
    Lfsr 1,USB_aRXReport                        ' Endpoint buffer
    Movlw cHID_INT_OUT_EP_SIZE                  ' Endpoint size
    Rcall __USB_EnableEndpointOut
'
' Enable and configure IN EndPoint
'
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Movlw cHID_EP
    Movwf GEN                                   ' Endpoint number
    Movlw cUSB_OUT_ENABLED | cUSB_IN_ENABLED | cUSB_HANDSHAKE_ENABLED | cUSB_DISALLOW_SETUP
    Movwf USB_wTemp                             ' Options
    Lfsr 1,USB_aTXReport                        ' Endpoint buffer
    Movlw cHID_INT_IN_EP_SIZE                   ' Endpoint size
    Bra __USB_EnableEndpointIn

'--------------------------------------------------------------------------------------------------------------------
' Subroutine:       __USB_CBSendResume
' Input:            None
' Output:           None
' Overview:         The USB specifications allow some types of USB peripheral devices to wake up a host PC (such as if it is in a low power suspend to RAM state).
'                   This can be a very useful feature in some USB applications,such as an Infrared remote control receiver.
'                   If a user presses the "power" button on a remote control,it is nice that the IR receiver can detect this signalling,and then send a USB "command" to the PC to wake up.
'
'                   The __USB_CBSendResume() "callback" subroutine is used to send this special USB signalling which wakes up the PC.
'                   This subroutine may be called by application firmware to wake up the PC.
'                   This subroutine should only be called when:
'
'                       1.  The USB driver used on the host PC supports the remote wakeup capability.
'                       2.  The USB configuration descriptor indicates the device is remote wakeup capable in the bmAttributes field.
'                       3.  The USB host PC is currently sleeping, and has previously sent your device a SET FEATURE setup packet which "armed" the remote wakeup capability.
'
'                   This callback should send a RESUME signal that has the period of 1-15ms.
'
' Notes:            Interrupt vs. Polling
'                   -Primary clock
'                   -Secondary clock
'                   > Can switch to primary first by calling __USB_CBWakeFromSuspend()

'                   The modifiable section in this routine should be changed to meet the application needs.
'                   Current implementation temporary blocks other functions from executing for a period of 1-13 ms depending on the core frequency.
'
'                   According to USB 2.0 specification section 7.1.7.7, "The remote wakeup device must hold the resume signaling
'                   for at least 1 ms but for no more than 15 ms."
'                   The idea here is to use a delay counter loop,using a common value that would work over a wide range of core frequencies.
'                   That value selected is 1800. See table below:
'                   ==========================================================
'                   Core Freq(MHz)      MIP         RESUME Signal Period (ms)
'                   ==========================================================
'                       48              12          1.05
'                        4              1           12.6
'                   ==========================================================
'                   These timing could be incorrect when using code optimisation or extended instruction mode, or when having other interrupts enabled.
'                   Make sure to verify using the MPLAB SIM's Stopwatch and verify the actual signal on an oscilloscope.
'
__USB_CBSendResume
    Movlb (_usb_ram >> 8)                       ' Point to the USB RAM bank
    Bsf UCON,Resume                             ' Start RESUME signaling
    Movlw 0x10                                  ' Set RESUME line for 1-13 ms
    Movwf FSR2H                                 ' Using FSR2 as temp
    Clrf FSR2L
USBCBSendResumeLoop
    Decfsz FSR2L,f
    Bra USBCBSendResumeLoop
    Decfsz FSR2H,f
    Bra USBCBSendResumeLoop
    Bcf UCON,Resume
    Return
Endasm-

$endif ' _USB_HID_INC_
