$ifndef _USB_MEM_INC_
$define _USB_MEM_INC_
'
' Memory allocation of USB routines
'
' Buffer Descriptor Byte Aliases
'
    Symbol cUSB_STAT   = 0              ' Buffer Descriptor Status Register
    Symbol cUSB_CNT    = 1              ' Buffer Count
    Symbol cUSB_ADRL   = 2              ' Buffer Address Low
    Symbol cUSB_ADRH   = 3              ' Buffer Address High
'
' STAT Bit Aliases
'
    Symbol cUSB_BSTALL = 2              ' Buffer Stall Enable
    Symbol cUSB_DTSEN  = 3              ' Data Toggle Synch Enable
    Symbol cUSB_INCDIS = 4              ' Address Increment Disable
    Symbol cUSB_KEN    = 5              ' BD Keep Enable
    Symbol cUSB_DTS    = 6              ' Data Toggle Synch Value
    Symbol cUSB_UOWN   = 7              ' USB Ownership
    Symbol cUSB_PID    = 2              ' Packet Identifier (4 bits)

    Dim GEN As Byte System              ' Used to hold the EndPoint number
    Dim PP1 As Byte System
    Dim PP1H As Byte System
    Dim USB_wSaveSFR As PP1.Word
'
' USB states
'
    Dim USB_bStatus As Byte System
    Dim USB_tConnected As USB_bStatus.0
    Dim USB_tPowered As USB_bStatus.1
    Dim USB_tDefault As USB_bStatus.2
    Dim USB_tPending As USB_bStatus.3
    Dim USB_tAddress As USB_bStatus.4
    Dim USB_tConfigured As USB_bStatus.5

    Dim USB_wFSR0 As FSR0L.Word
    Dim USB_wFSR1 As FSR1L.Word
    Dim USB_wFSR2 As FSR2L.Word
    Dim USB_wTBLPTR As TBLPTRL.Word

    Dim USB_tReady As STATUS.0          ' Set if we don't have control of the USB buffer
'
' Buffer Descriptor Table (BDT) RAM
'
' Endpoint 0, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 0)
    Dim USB_dEp0Bo As Dword At ( _usb_ram + 4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram + 4 * 1)

    Dim USB_wTemp As Word At USB_dEp0Bi + 4             ' Temporary storage
$endif
'
' Endpoint 1, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 1)
    Dim USB_dEp0Bo As Dword At ( _usb_ram + 4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram + 4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram + 4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)

    Dim USB_wTemp As Word At USB_dEp1Bi + 4             ' Temporary storage
$endif
'
' Endpoint 2, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 2)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)

    Dim USB_wTemp As Word At USB_dEp2Bi + 4             ' Temporary storage
$endif
'
' Endpoint 3, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 3)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)

    Dim USB_wTemp As Word At USB_dEp3Bi + 4             ' Temporary storage
$endif
'
' Endpoint 4, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 4)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)

    Dim USB_wTemp As Word At USB_dEp4Bi + 4             ' Temporary storage
$endif
'
' Endpoint 5, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 5)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)

    Dim USB_wTemp As Word At USB_dEp5Bi + 4             ' Temporary storage
$endif
'
' Endpoint 6, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 6)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)

    Dim USB_wTemp As Word At USB_dEp6Bi + 4             ' Temporary storage
$endif
'
' Endpoint 7, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 7)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)

    Dim USB_wTemp As Word At USB_dEp7Bi + 4             ' Temporary storage
$endif
'
' Endpoint 8, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 8)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)

    Dim USB_wTemp As Word At USB_dEp8Bi + 4             ' Temporary storage
$endif
'
' Endpoint 9, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 9)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)

    Dim USB_wTemp As Word At USB_dEp9Bi + 4             ' Temporary storage
$endif
'
' Endpoint 10, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 10)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)

    Dim USB_wTemp As Word At USB_dEp10Bi + 4            ' Temporary storage
$endif
'
' Endpoint 11, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 11)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)
    Dim USB_dEp11Bo As Dword At ( _usb_ram +  4 * 22)
    Dim USB_dEp11Bi As Dword At ( _usb_ram +  4 * 23)

    Dim USB_wTemp As Word At USB_dEp11Bi + 4            ' Temporary storage
$endif
'
' Endpoint 12, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 12)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)
    Dim USB_dEp11Bo As Dword At ( _usb_ram +  4 * 22)
    Dim USB_dEp11Bi As Dword At ( _usb_ram +  4 * 23)
    Dim USB_dEp12Bo As Dword At ( _usb_ram +  4 * 24)
    Dim USB_dEp12Bi As Dword At ( _usb_ram +  4 * 25)

    Dim USB_wTemp As Word At USB_dEp12Bi + 4            ' Temporary storage
$endif
'
' Endpoint 13, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 13)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)
    Dim USB_dEp11Bo As Dword At ( _usb_ram +  4 * 22)
    Dim USB_dEp11Bi As Dword At ( _usb_ram +  4 * 23)
    Dim USB_dEp12Bo As Dword At ( _usb_ram +  4 * 24)
    Dim USB_dEp12Bi As Dword At ( _usb_ram +  4 * 25)
    Dim USB_dEp13Bo As Dword At ( _usb_ram +  4 * 26)
    Dim USB_dEp13Bi As Dword At ( _usb_ram +  4 * 27)

    Dim USB_wTemp As Word At USB_dEp13Bi + 4            ' Temporary storage
$endif
'
' Endpoint 14, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 14)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)
    Dim USB_dEp11Bo As Dword At ( _usb_ram +  4 * 22)
    Dim USB_dEp11Bi As Dword At ( _usb_ram +  4 * 23)
    Dim USB_dEp12Bo As Dword At ( _usb_ram +  4 * 24)
    Dim USB_dEp12Bi As Dword At ( _usb_ram +  4 * 25)
    Dim USB_dEp13Bo As Dword At ( _usb_ram +  4 * 26)
    Dim USB_dEp13Bi As Dword At ( _usb_ram +  4 * 27)
    Dim USB_dEp14Bo As Dword At ( _usb_ram +  4 * 28)
    Dim USB_dEp14Bi As Dword At ( _usb_ram +  4 * 29)

    Dim USB_wTemp As Word At USB_dEp14Bi + 4            ' Temporary storage
$endif
'
' endpoint 15, OUT and IN
$if (cUSB_MAX_EP_NUMBER = 15)
    Dim USB_dEp0Bo As Dword At ( _usb_ram +  4 * 0)
    Dim USB_dEp0Bi As Dword At ( _usb_ram +  4 * 1)
    Dim USB_dEp1Bo As Dword At ( _usb_ram +  4 * 2)
    Dim USB_dEp1Bi As Dword At ( _usb_ram +  4 * 3)
    Dim USB_dEp2Bo As Dword At ( _usb_ram +  4 * 4)
    Dim USB_dEp2Bi As Dword At ( _usb_ram +  4 * 5)
    Dim USB_dEp3Bo As Dword At ( _usb_ram +  4 * 6)
    Dim USB_dEp3Bi As Dword At ( _usb_ram +  4 * 7)
    Dim USB_dEp4Bo As Dword At ( _usb_ram +  4 * 8)
    Dim USB_dEp4Bi As Dword At ( _usb_ram +  4 * 9)
    Dim USB_dEp5Bo As Dword At ( _usb_ram +  4 * 10)
    Dim USB_dEp5Bi As Dword At ( _usb_ram +  4 * 11)
    Dim USB_dEp6Bo As Dword At ( _usb_ram +  4 * 12)
    Dim USB_dEp6Bi As Dword At ( _usb_ram +  4 * 13)
    Dim USB_dEp7Bo As Dword At ( _usb_ram +  4 * 14)
    Dim USB_dEp7Bi As Dword At ( _usb_ram +  4 * 15)
    Dim USB_dEp8Bo As Dword At ( _usb_ram +  4 * 16)
    Dim USB_dEp8Bi As Dword At ( _usb_ram +  4 * 17)
    Dim USB_dEp9Bo As Dword At ( _usb_ram +  4 * 18)
    Dim USB_dEp9Bi As Dword At ( _usb_ram +  4 * 19)
    Dim USB_dEp10Bo As Dword At ( _usb_ram +  4 * 20)
    Dim USB_dEp10Bi As Dword At ( _usb_ram +  4 * 21)
    Dim USB_dEp11Bo As Dword At ( _usb_ram +  4 * 22)
    Dim USB_dEp11Bi As Dword At ( _usb_ram +  4 * 23)
    Dim USB_dEp12Bo As Dword At ( _usb_ram +  4 * 24)
    Dim USB_dEp12Bi As Dword At ( _usb_ram +  4 * 25)
    Dim USB_dEp13Bo As Dword At ( _usb_ram +  4 * 26)
    Dim USB_dEp13Bi As Dword At ( _usb_ram +  4 * 27)
    Dim USB_dEp14Bo As Dword At ( _usb_ram +  4 * 28)
    Dim USB_dEp14Bi As Dword At ( _usb_ram +  4 * 29)
    Dim USB_dEp15Bo As Dword At ( _usb_ram +  4 * 30)
    Dim USB_dEp15Bi As Dword At ( _usb_ram +  4 * 31)

    Dim USB_wTemp As Word At USB_dEp15Bi + 4                                                ' Temporary storage
$endif
'
' General RAM (in USB RAM)
'
    Dim USB_wSrc As Word At USB_wTemp + 2                                                   ' Source pointer
    Dim USB_wDst As Word At USB_wSrc + 2                                                    ' Destination pointer
    Dim USB_bDeviceState As Byte At USB_wDst + 2                                            ' Device State Notes :  DETACHED,ATTACHED,...
    Dim USB_bActiveConfiguration As Byte At USB_bDeviceState + 1                            ' Value of current configuration
    Dim USB_aAlternateInterface[cUSB_MAX_NUM_INT] As Byte At USB_bActiveConfiguration + 1   ' Array to keep track of the current alternate setting for each interface ID
    Dim USB_bControlTransferState As Byte At USB_aAlternateInterface + cUSB_MAX_NUM_INT     ' Control Transfer State
    Dim USB_wInCount As Word At USB_bControlTransferState + 1                               ' In data counter
    Dim USB_wOutCount As Word At USB_wInCount + 2                                           ' Out data counter
    Dim USB_bShortPacketStatus As Byte At USB_wOutCount + 2                                 ' Flag used by Control Transfer Read
    Dim USB_bUSTATcopy As Byte At USB_bShortPacketStatus + 1
    Dim USB_bLastDTS As Byte At USB_bUSTATcopy + 1
    Dim USB_bTRNIFCount As Byte At USB_bLastDTS + 1
    Dim USB_bInFo As Byte At USB_bTRNIFCount + 1
    Dim USB_bOutFo As Byte At USB_bInFo + 1
    Dim USB_bRemoteWakeup As Byte At USB_bOutFo + 1
'
' USB_bInFo bit aliases
'
    Symbol cUSB_CtrlTrfMem  = 0                     ' [0]RAM [1]ROM
    Symbol cUSB_IncludeZero = 6
    Symbol cUSB_Busy        = 7
'
' Endpoint 0 buffers
'
    Dim USB_aSetupPkt[cUSB_EP0_BUFF_SIZE] As Byte At USB_bRemoteWakeup + 1
    Dim USB_aCtrlTrfData[cUSB_EP0_BUFF_SIZE] As Byte At USB_aSetupPkt + cUSB_EP0_BUFF_SIZE
'
' USB_aSetupPkt Byte and Bit Aliases
'
    Symbol cUSB_bmRequestType   = 0
    Symbol cUSB_bRequest        = 1
    Symbol cUSB_wValue          = 2
    Symbol cUSB_wIndex          = 4
    Symbol cUSB_wLength         = 6

    Symbol cUSB_Recipient       = 0         ' Device,Interface,Endpoint,Other (5 bits)
    Symbol cUSB_RequestType     = 5         ' Standard,Class,Vendor,Reserved (2 bits)
    Symbol cUSB_DataDir         = 7         ' Host-to-device,Device-to-host
    Symbol cUSB_bFeature        = 2         ' cDEVICE_REMOTE_WAKEUP,cENDPOINT_HALT

    Symbol cUSB_bReportID       = 2         ' In a cSET_REPORT or cGET_REPORT request
    Symbol cUSB_bReportType     = 3         ' In a cSET_REPORT or cGET_REPORT request

    Symbol cUSB_bDscIndex       = 2         ' For Configuration and String DSC Only
    Symbol cUSB_bDescriptorType = 3         ' Device,Configuration,String
    Symbol cUSB_wLangID         = 4         ' Language ID

    Symbol cUSB_bDevADR         = 2         ' Device Address 0-127
    Symbol cUSB_bDevADRH        = 3         ' Must equal zero

    Symbol cUSB_bConfigValue    = 2         ' Configuration Value 0-255
    Symbol cUSB_bCfgRSD         = 3         ' Must equal zero (Reserved)

    Symbol cUSB_bAltID          = 2         ' Alternate Setting Value 0-255
    Symbol cUSB_bAltIDH         = 3         ' Must equal zero (Reserved)
    Symbol cUSB_bIntfID         = 4         ' Interface Number Value 0-255
    Symbol cUSB_bIntfIDH        = 5         ' Must equal zero

    Symbol cUSB_bEPID           = 4         ' Endpoint ID (Number & Direction)
    Symbol cUSB_bEPIDH          = 5         ' Must equal zero

    Symbol cUSB_EPNum           = 0         ' Endpoint Number 0-15 (4 bits)
    Symbol cUSB_EPDir           = 7         ' Endpoint Direction :  0-OUT,1-IN
'
' HID RAM (in _usb_ram bank)
'
$ifdef cUSB_USE_HID
    Dim USB_bIdleRate As Byte At USB_aCtrlTrfData + cUSB_EP0_BUFF_SIZE
    Dim USB_bActiveProtocol As Byte At USB_bIdleRate + 1
'
' HID Buffer RAM
'
    Dim USB_aRXReport[cHID_INT_OUT_EP_SIZE + 1] As Byte At USB_bActiveProtocol + 1
    Dim USB_aTXReport[cHID_INT_IN_EP_SIZE + 1] As Byte At USB_aRXReport + cHID_INT_OUT_EP_SIZE + 1
    Dim USB_dRXReport As Dword At USB_aRXReport#1
    Dim USB_dTXReport As Dword At USB_aTXReport#1
    Dim __USBOut_Buffer As String * cHID_INT_OUT_EP_SIZE At USB_aRXReport#1
    Dim __USBin_Buffer As String * cHID_INT_IN_EP_SIZE At USB_aTXReport#1
$endif
'
' CDC RAM (in _usb_ram bank)
'
$ifdef cUSB_USE_CDC
    Dim USB_aLineCoding[cLINE_CODING_LENGTH] As Byte At USB_aCtrlTrfData + cUSB_EP0_BUFF_SIZE
    Dim USB_bControlSignalBitmap As Byte At USB_aLineCoding + cLINE_CODING_LENGTH
    Dim USB_aDummyEncapCmdResp[cDUMMY_LENGTH] As Byte At USB_bControlSignalBitmap + 1
'
' CDC Buffer RAM
'
    Dim USB_aCdcNotice[cCDC_COMM_IN_EP_SIZE + 1] As Byte At USB_aDummyEncapCmdResp + cDUMMY_LENGTH
    Dim USB_aRXReport[cCDC_DATA_OUT_EP_SIZE + 1] As Byte At USB_aCdcNotice + cCDC_DATA_OUT_EP_SIZE + 1
    Dim USB_aTXReport[cCDC_DATA_IN_EP_SIZE + 1] As Byte At USB_aRXReport + cCDC_DATA_IN_EP_SIZE + 1
    Dim USB_dRXReport As Dword At USB_aRXReport#1
    Dim USB_dTXReport As Dword At USB_aTXReport#1
    Dim __USBOut_Buffer As String * cCDC_DATA_OUT_EP_SIZE At USB_aRXReport#1
    Dim __USBin_Buffer As String * cCDC_DATA_IN_EP_SIZE At USB_aTXReport#1
'
' USB_aLineCoding Byte and Bit Aliases
'
    $define cUSB_dwDTERate   0                           ' Complex data structure
    $define cUSB_bCharFormat 4
    $define cUSB_bParityType 5
    $define cUSB_bDataBits   6
'
' USB_bControlSignalBitmap Byte and Bit Aliases
'
    $define cUSB_DTE_PRESENT     0                       ' [0] Not Present [1] Present
    $define cUSB_CARRIER_CONTROL 1                       ' [0] Deactivate  [1] Activate
$endif

$endif ' _USB_MEM_INC_
